/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000-2003 Sakai Hiroaki.                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* Canon 饹(ˤѤΥ饹)Υ饤֥                        */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "CanonP.h"

/*****************************************************************************/
/* ؿ                                                                */
/*****************************************************************************/

/*===========================================================================*/
/* ѿμ                                                          */
/*===========================================================================*/

double CanonGetX( Canon canon) { return (canon->x ); }
double CanonGetY( Canon canon) { return (canon->y ); }
double CanonGetVX(Canon canon) { return (canon->vx); }
double CanonGetVY(Canon canon) { return (canon->vy); }
int CanonGetState(Canon canon)        { return (canon->state); }
int CanonSetState(Canon canon, int s) { return (canon->state = s); }
Tank CanonGetTank(Canon canon)         { return (canon->tank); }
Tank CanonSetTank(Canon canon, Tank tank)
{
  canon->tank = tank;
  if (canon->tank) TankIncCanon(canon->tank);
  return (canon->tank);
}

Tank CanonChangeTank(Canon canon, Tank tank)
{
  if (canon->tank) TankDecCanon(canon->tank);
  return (CanonSetTank(canon, tank));
}

/*===========================================================================*/
/* ȯȽ                                                                */
/*===========================================================================*/

static int CanonCheckExplosion(Canon canon, Map map, ObjList tank_list,
			       ObjList explosion_list)
{
  Tank tank;
  Explosion explosion;
  double x, y, d, size;
  ObjListData current;

  if (MapGetState(map,
		  ConvertXFromPositionToMap(canon->x),
		  ConvertYFromPositionToMap(canon->y))) {
#if 0 /* ޥåץǡιϡExplosion.c ֥ȤԤ */
    MapChangeState(map,
		   ConvertXFromPositionToMap(canon->x),
		   ConvertYFromPositionToMap(canon->y), 0);
#endif
    canon->state = CANON_EXPLOSION;
    return (canon->state);
  }

  size = (double)EXPLOSION_SIZE / CHARACTER_SIZE * 0.5 * 1.2;

  for (current = ObjList_GetStart(tank_list);
       !ObjList_IsEndEdge(tank_list, current);
       current = ObjListData_GetNext(current)) {
    tank = (Tank)ObjListData_GetObj(current);
    x = canon->x - TankGetX(tank);
    y = canon->y - TankGetY(tank);
    d = x * x + y * y;
    if (d < size * size) {
      canon->state = CANON_EXPLOSION;
      return (canon->state);
    }
  }

  for (current = ObjList_GetStart(explosion_list);
       !ObjList_IsEndEdge(explosion_list, current);
       current = ObjListData_GetNext(current)) {
    explosion = (Explosion)ObjListData_GetObj(current);
    if (ExplosionCheck(explosion, canon->x, canon->y)) {
      canon->state = CANON_EXPLOSION;
      return (canon->state);
    }
  }

  return (canon->state);
}

/*===========================================================================*/
/* ֥ȤȺ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

static Canon CanonAllocate(double x, double y, double vx, double vy, Tank tank,
			   int state)
{
  Canon canon;

  canon = (Canon)malloc(sizeof(_Canon));
  if (canon == NULL) ErrorExit("Cannot create canon.");

  canon->x = x + vx * CANON_OFFSET;
  canon->y = y + vy * CANON_OFFSET;
  canon->vx = vx;
  canon->vy = vy;
  canon->tank = tank;
  canon->state = state;

  return (canon);
}

Canon CanonCreate(double x, double y, double vx, double vy, Tank tank,
		  int state)
{
  Canon canon;

  canon = CanonAllocate(x, y, vx, vy, tank, state);
  if (canon == NULL) ErrorExit("Cannot create canon.");
  if (canon->tank) TankIncCanon(canon->tank);

  return (canon);
}

Canon CanonCreateFromCanon(Canon old_canon)
{
  Canon canon;

  canon = CanonAllocate(old_canon->x, old_canon->y,
			old_canon->vx, old_canon->vy,
			old_canon->tank, old_canon->state);
  if (canon == NULL) ErrorExit("Cannot create canon.");

  /*
   * ԡƺˤϡ󥯥ȤԤʤ
   * ( CanonSetTank() ꤹ٤)
   *  destroy ˤϥǥȤԤΤա
   * (Τޤ destroy ȡ礬)
   */
  /* if (canon->tank) TankIncCanon(canon->tank); */

  return (canon);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

Canon CanonDestroy(Canon canon)
{
  if (canon == NULL) return (NULL);

  if (canon->tank) TankDecCanon(canon->tank);
  free(canon);

  return (NULL);
}

/*===========================================================================*/
/* ư                                                                  */
/*===========================================================================*/

static int CanonMove(Canon canon)
{
  canon->x += canon->vx;
  canon->y += canon->vy;

  return (0);
}

int CanonAct(Canon canon, Map map, ObjList tank_list, ObjList explosion_list)
{
  switch (canon->state) {
  case CANON_WAIT :
    break;
  case CANON_ACTIVE :
    CanonMove(canon);
    CanonCheckExplosion(canon, map, tank_list, explosion_list);
    break;
  case CANON_EXPLOSION :
    break;
  case CANON_DESTROY :
    break;
  default :
    break;
  }
  return (canon->state);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
