/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000-2003 Sakai Hiroaki.                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* Enemy 饹(ŨΥ饹)Υ饤֥                                */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "EnemyP.h"

/*****************************************************************************/
/* ؿ                                                                */
/*****************************************************************************/

/*===========================================================================*/
/* ѿμ                                                    */
/*===========================================================================*/

int EnemyGetType(Enemy e)           { return (e->type       ); }
int EnemySetType(Enemy e, int type) { return (e->type = type); }

/*===========================================================================*/
/* Ũΰư                                                                  */
/*===========================================================================*/

static int calc_abs(int x)
{
  return (x < 0 ? -x : x);
}

static int calc_range(double x0, double y0, double x1, double y1, int r,
		      int width, int height, int peeping)
{
  double dx, dy;
  dx = x0 - x1;
  dy = y0 - y1;
  if (peeping) {
    if ((calc_abs(dx) > width / 2) || (calc_abs(dy) > height / 2)) {
      return (0);
    }
  }
  if (!r) return (1);
  if (dx * dx + dy * dy <= r * r) {
    return (1);
  }
  return (0);
}

/* angle ϣ360٤ǻ */
static int calc_angle(Tank enemy_tank, Tank target_tank, int angle)
{
  int degree, tank_angle;

  if (enemy_tank == NULL) return (0);
  if (target_tank == NULL) return (0);

  degree = easy_atan(TankGetX(target_tank) - TankGetX(enemy_tank),
		     TankGetY(target_tank) - TankGetY(enemy_tank));
  degree -= angle;
  tank_angle = (degree * TANK_ANGLE_NUMBER / 360) - TankGetAngle(enemy_tank);

  /* -180 180ϰϤĴ᤹ */
  while (tank_angle < 0) {
    tank_angle += TANK_ANGLE_NUMBER;
  }
  tank_angle += TANK_ANGLE_NUMBER / 2;
  tank_angle %= TANK_ANGLE_NUMBER;
  tank_angle -= TANK_ANGLE_NUMBER / 2;

  return (tank_angle);
}

/* angle ϣ360٤ǻ */
static unsigned long int turn_to_target(Tank enemy_tank, Tank target_tank,
					int angle)
{
  int tank_angle;

  if (enemy_tank == NULL) return (TANK_NO_ACTION);
  if (target_tank == NULL) return (TANK_NO_ACTION);
  tank_angle = calc_angle(enemy_tank, target_tank, angle);
  if (tank_angle < 0) return (TANK_TURN_LEFT);
  if (tank_angle > 0) return (TANK_TURN_RIGHT);
  return (TANK_NO_ACTION);
}

#define ESCAPE_FLAG (1 << 0)
#define SHOT_FLAG (1 << 1)

static int calc_future(unsigned long int flags,
		       Enemy enemy, Map map,
		       ObjList tank_list_orig, ObjList canon_list_orig,
		       ObjList missile_list_orig, ObjList explosion_list_orig,
		       int width, int height, int peeping,
		       int range, int depth,
		       int * future_step,
		       int * moved,
		       int * moved_distance,
		       int * target_distance,
		       int * target_distance_delta,
		       int * target_angle)
{
  ObjList tank_list;
  ObjList canon_list;
  ObjList missile_list;
  ObjList explosion_list;
  ObjListData current, current2;
  Tank tank_orig, tank;
  Canon canon_orig, canon;
  Missile missile_orig, missile;
  Explosion explosion_orig, explosion;
  Tank enemy_tank, target_tank;
  int i;
  int ret_value = -1;
  double dx, dy;
  double enemy_old_x, enemy_old_y;
  double target_distance_old;

  MapUndo(map);

  tank_list = ObjList_Create();
  canon_list = ObjList_Create();
  missile_list = ObjList_Create();
  explosion_list = ObjList_Create();

  enemy_tank = PlayerGetTank((Player)enemy);
  target_tank = enemy->target;

  enemy_old_x = TankGetX(enemy_tank);
  enemy_old_y = TankGetY(enemy_tank);

  if (future_step)
    *future_step = -1;

  if (moved) *moved = 0;
  if (moved_distance) {
    *moved_distance = 0;
  }
  target_distance_old = 0.0;
  if (target_distance_delta) {
    *target_distance_delta = 0;
    if (target_tank) {
      dx = enemy_old_x - TankGetX(target_tank);
      dy = enemy_old_y - TankGetY(target_tank);
      target_distance_old = easy_sqrt(dx * dx + dy * dy);
    }
  }

  if (canon_list_orig) {
    for (current = ObjList_GetStart(canon_list_orig);
	 !ObjList_IsEndEdge(canon_list_orig, current);
	 current = ObjListData_GetNext(current)) {
      canon_orig = (Canon)ObjListData_GetObj(current);
      if (CanonGetState(canon_orig) != CANON_ACTIVE) continue;
      if (!calc_range(TankGetX(enemy_tank), TankGetY(enemy_tank),
		      CanonGetX(canon_orig), CanonGetY(canon_orig), range,
		      width, height, peeping))
	continue;
      canon = CanonCreateFromCanon(canon_orig);
      ObjList_InsertObjToEnd(canon_list, canon,
			     (ObjDestructor)CanonDestroy);
    }
  }

  if (missile_list_orig) {
    for (current = ObjList_GetStart(missile_list_orig);
	 !ObjList_IsEndEdge(missile_list_orig, current);
	 current = ObjListData_GetNext(current)) {
      missile_orig = (Missile)ObjListData_GetObj(current);
      if (MissileGetState(missile_orig) != MISSILE_ACTIVE) continue;
      if (!calc_range(TankGetX(enemy_tank), TankGetY(enemy_tank),
		      MissileGetX(missile_orig), MissileGetY(missile_orig),
		      range, width, height, peeping))
	continue;
      missile = MissileCreateFromMissile(missile_orig);
      ObjList_InsertObjToEnd(missile_list, missile,
			     (ObjDestructor)MissileDestroy);
    }
  }

  if (explosion_list_orig) {
    for (current = ObjList_GetStart(explosion_list_orig);
	 !ObjList_IsEndEdge(explosion_list_orig, current);
	 current = ObjListData_GetNext(current)) {
      explosion_orig = (Explosion)ObjListData_GetObj(current);
      if (ExplosionGetState(explosion_orig) != EXPLOSION_ACTIVE) continue;
      if (!calc_range(TankGetX(enemy_tank), TankGetY(enemy_tank),
		      ExplosionGetX(explosion_orig),
		      ExplosionGetY(explosion_orig),
		      range, width, height, peeping))
	continue;
      explosion = ExplosionCreateFromExplosion(explosion_orig);
      ObjList_InsertObjToEnd(explosion_list, explosion,
			     (ObjDestructor)ExplosionDestroy);
    }
  }

  if (!tank_list_orig) {
    /*
     * ֤¾Υ֥Ȥ黲ȤƤΤǡNULL ϤǤʤ
     * ɬԡʤФʤʤ
     */
    ErrorExit("Tank list is NULL.");
  }

  for (current = ObjList_GetStart(tank_list_orig);
       !ObjList_IsEndEdge(tank_list_orig, current);
       current = ObjListData_GetNext(current)) {
    tank_orig = (Tank)ObjListData_GetObj(current);
    /*
     * ֤äˤƤȤޤĤäƤ꥿åȤˤʤäƤ뤫
     * ʤΤǡACTIVE ǤʤƤ⥹åפǤʤ
     */
    /* if (TankGetState(tank_orig) != TANK_ACTIVE) continue; */
    tank = TankCreateFromTank(tank_orig);
    ObjList_InsertObjToEnd(tank_list, tank,
			   (ObjDestructor)TankDestroy);

    /*
     * ץ쥤䡼ξ󤬸äƽ񤭴뤳ȤΤʤ褦
     * NULL ꤷƤ
     */
    TankSetPlayer(tank, NULL);

    if (enemy_tank == tank_orig) {
      enemy_tank = tank;
    } else {
      /*
       * ŨϾˡ¨¤˼ͷ⤷ƤΤȲꤹ롥
       * ȤŨФƲˤʤäƤȤȤ꤬㳲ʪ餱
       * ưʤȤ˷⤿ȡȤɡ褦ʤΤǡ餫
       * Τ褦ʾ֤ˤʤʤ褦ˤ뤿н补
       */
      TankSetAction(tank, TankGetAction(tank) | TANK_FIRE_CANON);
    }
    if (target_tank == tank_orig) {
      target_tank = tank;
    }

    for (current2 = ObjList_GetStart(canon_list);
	 !ObjList_IsEndEdge(canon_list, current2);
	 current2 = ObjListData_GetNext(current2)) {
      canon = (Canon)ObjListData_GetObj(current2);

      if (CanonGetTank(canon) == tank_orig) {
	/* ˤƤοƤѹ */
	CanonSetTank(canon, tank);
      }
    }

    for (current2 = ObjList_GetStart(missile_list);
	 !ObjList_IsEndEdge(missile_list, current2);
	 current2 = ObjListData_GetNext(current2)) {
      missile = (Missile)ObjListData_GetObj(current2);

      if (MissileGetTank(missile) == tank_orig) {
	/* ߥοƤѹ */
	MissileSetTank(missile, tank);
      }
      if (MissileGetTarget(missile) == tank_orig) {
	/* ߥΥåȤѹ */
	MissileSetTarget(missile, tank);
      }
    }

    for (current2 = ObjList_GetStart(explosion_list);
	 !ObjList_IsEndEdge(explosion_list, current2);
	 current2 = ObjListData_GetNext(current2)) {
      explosion = (Explosion)ObjListData_GetObj(current2);

      if (ExplosionGetTank(explosion) == tank_orig) {
	/* ȯοƤѹ */
	ExplosionSetTank(explosion, tank);
      }
    }
  }

  for (i = 0; i < depth; i++) {

    if (enemy->actions_num > enemy->current_action + i) {
      TankSetAction(enemy_tank, enemy->actions[enemy->current_action + i]);
    } else {
#if 0 /* ꤷʤ(Υ³Ѥ) */
      TankSetAction(enemy_tank, TANK_NO_ACTION);
#endif
    }

    /* ֤ν */
    StageProcTank(map, tank_list, canon_list, missile_list, explosion_list);
    if (moved)
      *moved += TankGetMoved(enemy_tank);

    if (flags & ESCAPE_FLAG) {
      if (TankGetState(enemy_tank) != TANK_ACTIVE) {
	ret_value = i;
	goto ret;
      }
    }

    if (flags & SHOT_FLAG) {
      for (current = ObjList_GetStart(tank_list);
	   !ObjList_IsEndEdge(tank_list, current);
	   current = ObjListData_GetNext(current)) {
	tank = (Tank)ObjListData_GetObj(current);
	if (tank == enemy_tank) continue;
	if (TankGetState(tank) != TANK_ACTIVE) {
	  if (TankGetDestroyed(tank) == enemy_tank) {
	    ret_value = i;
	    goto ret;
	  }
	}
      }
    }

    /* ȯν */
    StageProcExplosion(map, explosion_list);

    /* ˤƤν */
    StageProcCanon(map, tank_list, canon_list, explosion_list);

    /* ߥν */
    StageProcMissile(map, tank_list, missile_list, explosion_list);

    for (current = ObjList_GetStart(canon_list);
	 !ObjList_IsEndEdge(canon_list, current);
	 current = ObjListData_GetNext(current)) {
      canon = (Canon)ObjListData_GetObj(current);
      if (CanonGetState(canon) != CANON_ACTIVE) continue;
      if (!calc_range(TankGetX(enemy_tank), TankGetY(enemy_tank),
		      CanonGetX(canon), CanonGetY(canon), range,
		      width, height, peeping)) {
	current = ObjList_DeleteObjToPrev(canon_list, current);
      }
    }

    for (current = ObjList_GetStart(missile_list);
	 !ObjList_IsEndEdge(missile_list, current);
	 current = ObjListData_GetNext(current)) {
      missile = (Missile)ObjListData_GetObj(current);
      if (MissileGetState(missile) != MISSILE_ACTIVE) continue;
      if (!calc_range(TankGetX(enemy_tank), TankGetY(enemy_tank),
		      MissileGetX(missile), MissileGetY(missile), range,
		      width, height, peeping)) {
	current = ObjList_DeleteObjToPrev(missile_list, current);
      }
    }
  }

ret:

  if (future_step)
    *future_step = ret_value;

  /* depth Фɽ */
  if (ret_value >= 0) {
    ret_value = depth ? ((ret_value + 1) * 100 / depth) : 0;
  }

  /* ư̤ depth Фɽ */
  if (moved) {
    /* moved 1 4 Τǡ4 ǳäƤ */
    *moved = depth ? (*moved * 100 / (4 * depth)) : 0;
  }

  /* ưΥ depth Фɽ */
  if (moved_distance) {
    if (depth > 0) {
    dx = TankGetX(enemy_tank) - enemy_old_x;
    dy = TankGetY(enemy_tank) - enemy_old_y;
    *moved_distance =
      (int)(easy_sqrt(dx * dx + dy * dy) / (depth * TANK_SPEED) * 100);
    } else {
      *moved_distance = 0;
    }
  }

  /*
   * åȤȤεΥ
   */
  if (target_distance) {
    if (target_tank != NULL) {
      dx = TankGetX(enemy_tank) - TankGetX(target_tank);
      dy = TankGetY(enemy_tank) - TankGetY(target_tank);
      *target_distance = (int)easy_sqrt(dx * dx + dy * dy);
    } else {
      *target_distance = 0;
    }
  }

  /*
   * åȤȤεΥѲ depth Фɽ
   * դ󤶤äͤˤʤ롥
   * åȤưΤǡ100 Ķ⤢롥
   */
  if (target_distance_delta) {
    if (target_tank != NULL && depth > 0) {
      dx = TankGetX(enemy_tank) - TankGetX(target_tank);
      dy = TankGetY(enemy_tank) - TankGetY(target_tank);
      *target_distance_delta =
	(int)((target_distance_old - easy_sqrt(dx * dx + dy * dy))
	      / (depth * TANK_SPEED) * 100);
    } else {
      *target_distance_delta = 0.0;
    }
  }

  /* Ⱦ100ʬȤ-100100 ɽ */
  if (target_angle) {
    *target_angle =
      (calc_angle(enemy_tank, target_tank, 0) * 200) / TANK_ANGLE_NUMBER;
  }

  ObjList_Destroy(canon_list);
  ObjList_Destroy(missile_list);
  ObjList_Destroy(explosion_list);

  /* ֤¾Υ֥ȤοƤˤʤΤǡǸ˾õ */
  ObjList_Destroy(tank_list);

  return (ret_value);
}

#define RANDOM_ACTION_NO_FLAG              0
#define RANDOM_ACTION_FORWARD_ONLY        (1 << 0)
#define RANDOM_ACTION_BACKWARD_ONLY       (1 << 1)
#define RANDOM_ACTION_FORWARD_OR_BACKWARD (1 << 2)
#define RANDOM_ACTION_USE_STAY            (1 << 3)
#define RANDOM_ACTION_TURN_TO_TARGET      (1 << 4)

static void set_random_action(Enemy enemy, int depth, int count,
			      int angle, unsigned long int flags)
{
  int i, j;
  unsigned long int actions[3][3] = {
    {
      TANK_GO_FORWARD,
      TANK_GO_FORWARD | TANK_TURN_LEFT,
      TANK_GO_FORWARD | TANK_TURN_RIGHT
    }, {
      TANK_GO_BACKWARD,
      TANK_GO_BACKWARD | TANK_TURN_LEFT,
      TANK_GO_BACKWARD | TANK_TURN_RIGHT
    }, {
      TANK_NO_ACTION,
      TANK_TURN_LEFT,
      TANK_TURN_RIGHT
    }
  };
  unsigned long int action;
  int go_or_back = 0;

  if (flags & RANDOM_ACTION_FORWARD_ONLY)
    go_or_back = 0;
  else if (flags & RANDOM_ACTION_BACKWARD_ONLY)
    go_or_back = 1;
  else if (flags & RANDOM_ACTION_FORWARD_OR_BACKWARD)
    go_or_back = GetRandomValue(2);

  enemy->actions_num = 0;
  enemy->current_action = 0;
  for (i = 0; i < count; i++) {
    if (!(flags & (RANDOM_ACTION_FORWARD_ONLY |
		   RANDOM_ACTION_BACKWARD_ONLY |
		   RANDOM_ACTION_FORWARD_OR_BACKWARD))) {
      if (flags & RANDOM_ACTION_USE_STAY)
	go_or_back = GetRandomValue(3);
      else
	go_or_back = GetRandomValue(2);
    }
    action = actions[go_or_back][GetRandomValue(3)];
    if (GetRandomValue(3) && (flags & RANDOM_ACTION_TURN_TO_TARGET)) {
      action &= ~(TANK_TURN_LEFT | TANK_TURN_RIGHT);
      action |=
	turn_to_target(PlayerGetTank((Player)enemy), enemy->target, angle)
	& (TANK_TURN_LEFT | TANK_TURN_RIGHT);
    }
    for (j = 0; j < depth; j++) {
      enemy->actions[enemy->actions_num] = action;
      (enemy->actions_num)++;
    }
  }
  return;
}

static int shot_canon_and_missile(Enemy enemy, Map map, 
				  ObjList tank_list, ObjList canon_list,
				  ObjList missile_list, ObjList explosion_list,
				  int width, int height, int peeping)
{
  int shot;

  if (TankGetCanonTimer(PlayerGetTank((Player)enemy)))
    return (0);

  if (enemy->actions_num <= enemy->current_action)
    return (0);

  if (enemy->actions[enemy->current_action] &
      (TANK_FIRE_CANON | TANK_FIRE_MISSILE))
    return (0);

#if 0
  if (GetRandomValue(3))
    return (0);
#endif

  enemy->actions[enemy->current_action] |= TANK_FIRE_CANON;
  shot = calc_future(SHOT_FLAG,
		     enemy, map, tank_list, canon_list,
		     missile_list, explosion_list,
		     width, height, peeping,
		     0, 20, NULL, NULL, NULL, NULL, NULL, NULL);
  if (shot >= 0)
    return (1);
  enemy->actions[enemy->current_action] &= ~TANK_FIRE_CANON;

  if (TankGetMissileTimer(PlayerGetTank((Player)enemy)))
    return (0);

  if (GetRandomValue(70))
    return (0);

  enemy->actions[enemy->current_action] |= TANK_FIRE_MISSILE;
  shot = calc_future(SHOT_FLAG,
		     enemy, map, tank_list, canon_list,
		     missile_list, explosion_list,
		     width, height, peeping,
		     0, 20, NULL, NULL, NULL, NULL, NULL, NULL);
  if (shot >= 0)
    return (1);
  enemy->actions[enemy->current_action] &= ~TANK_FIRE_MISSILE;

  return (0);
}

static int select_target(Enemy enemy, ObjList tank_list, int r,
			 int width, int height, int peeping)
{
  ObjListData current;
  Tank tank = NULL;
  double ex, ey;
  double x, y, dx, dy, d, min_d;
  int n;

  ex = TankGetX(PlayerGetTank((Player)enemy));
  ey = TankGetY(PlayerGetTank((Player)enemy));
  enemy->target = NULL;

  if (GetRandomValue(100) < r) {
    min_d = -1.0;
    for (current = ObjList_GetStart(tank_list);
	 !ObjList_IsEndEdge(tank_list, current);
	 current = ObjListData_GetNext(current)) {
      tank = (Tank)ObjListData_GetObj(current);
      if (TankGetState(tank) != TANK_ACTIVE) continue;
      if (tank == PlayerGetTank((Player)enemy)) continue;
      x = TankGetX(tank);
      y = TankGetY(tank);
      dx = ex - x; dy = ey - y;
      if (peeping) {
	if ((calc_abs(dx) > width / 2) || (calc_abs(dy) > height / 2))
	  continue;
      }
      d = dx * dx + dy * dy;
      if ((min_d < 0.0) || (min_d > d)) {
	enemy->target = tank;
	min_d = d;
      }
    }
  } else {
    n = 0;
    for (current = ObjList_GetStart(tank_list);
	 !ObjList_IsEndEdge(tank_list, current);
	 current = ObjListData_GetNext(current)) {
      tank = (Tank)ObjListData_GetObj(current);
      if (TankGetState(tank) != TANK_ACTIVE) continue;
      if (tank == PlayerGetTank((Player)enemy)) continue;
      x = TankGetX(tank);
      y = TankGetY(tank);
      dx = ex - x; dy = ey - y;
      if (peeping) {
	if ((calc_abs(dx) > width / 2) || (calc_abs(dy) > height / 2))
	  continue;
      }
      n++;
    }
    if (!n) {
      enemy->target = NULL;
      return (0);
    }
    n = GetRandomValue(n);
    for (current = ObjList_GetStart(tank_list);
	 !ObjList_IsEndEdge(tank_list, current);
	 current = ObjListData_GetNext(current)) {
      tank = (Tank)ObjListData_GetObj(current);
      if (TankGetState(tank) != TANK_ACTIVE) continue;
      if (tank == PlayerGetTank((Player)enemy)) continue;
      if (!n) break;
      x = TankGetX(tank);
      y = TankGetY(tank);
      dx = ex - x; dy = ey - y;
      if (peeping) {
	if ((calc_abs(dx) > width / 2) || (calc_abs(dy) > height / 2))
	  continue;
      }
      n--;
    }
    enemy->target = tank;
  }
  return (0);
}

static int calc_priority_danger(int future, int future_step,
				int moved, int moved_distance,
				int target_distance_delta, int angle)
{
  int ret = 0;

  /* ʤȤäƤʤΤͥ */
  ret += ((future < 0) ? 200 : future) * 1000;

  ret += moved * 2;
  ret += moved_distance * 2;

  /* ʤ٤ʤΥƤ */
  ret += (200 - target_distance_delta);

  return (ret);
}

static int calc_priority_escape(int future, int future_step,
				int moved, int moved_distance,
				int target_distance_delta, int angle)
{
  int ret = 0;

  /* ʤȤäƤʤΤͥ */
  ret += ((future < 0) ? 200 : future) * 1000;

  ret += moved * 2;
  ret += moved_distance * 2;

  /* ʤ٤ʤΥƤ */
  ret += (200 - target_distance_delta);

  return (ret);
}

static int calc_priority_near(int future, int future_step,
			      int moved, int moved_distance,
			      int target_distance_delta, int angle)
{
  int ret = 0;

  /* ʤȤäƤʤΤͥ */
  ret += ((future < 0) ? 200 : future) * 100;

  ret += moved;
  ret += moved_distance * 2;

  /* դƤ */
  ret += target_distance_delta * 2;

  /* åȤΤۤ */
  ret += (200 - calc_abs(angle)) * 4;

  return (ret);
}

static int calc_priority_nearest(int future, int future_step,
				 int moved, int moved_distance,
				 int target_distance_delta, int angle)
{
  int ret = 0;

  /* ʤȤäƤʤΤͥ */
  ret += ((future < 0) ? 200 : future) * 100;

  ret += moved;
  ret += moved_distance * 2;

  /* åȤΤۤ */
  ret += (200 - calc_abs(angle)) * 4;

  return (ret);
}

static int calc_priority_chase(int future, int future_step,
			       int moved, int moved_distance,
			       int target_distance_delta, int angle)
{
  int ret = 0;

  /* ʤȤäƤʤΤͥ */
  ret += ((future < 0) ? 200 : future) * 100;

  ret += moved;
  ret += moved_distance;

  /* դƤ */
  ret += target_distance_delta * 3;

  /* åȤΤۤ */
  ret += (200 - calc_abs(angle)) * 4;

  return (ret);
}

static int calc_priority_chase_left(int future, int future_step,
				    int moved, int moved_distance,
				    int target_distance_delta, int angle)
{
  int ret = 0;

  /* ʤȤäƤʤΤͥ */
  ret += ((future < 0) ? 200 : future) * 100;

  ret += moved;
  ret += moved_distance * 2;

  /* Υݤ */
  ret += (200 - calc_abs(target_distance_delta)) * 2;

  /* åȤκ */
  angle -= 50;
  if (angle < -100) angle += 200;
  ret += (200 - calc_abs(angle)) * 4;

  return (ret);
}

static int calc_priority_chase_right(int future, int future_step,
				     int moved, int moved_distance,
				     int target_distance_delta, int angle)
{
  int ret = 0;

  /* ʤȤäƤʤΤͥ */
  ret += ((future < 0) ? 200 : future) * 100;

  ret += moved;
  ret += moved_distance * 2;

  /* Υݤ */
  ret += (200 - calc_abs(target_distance_delta)) * 2;

  /* åȤα */
  angle += 50;
  if (angle > 100) angle -= 200;
  ret += (200 - calc_abs(angle)) * 4;

  return (ret);
}

static int calc_priority_back(int future, int future_step,
			      int moved, int moved_distance,
			      int target_distance_delta, int angle)
{
  int ret = 0;

  /* ʤȤäƤʤΤͥ */
  ret += ((future < 0) ? 200 : future) * 100;

  ret += moved;
  ret += moved_distance;

  /* ʤ٤ʤΥƤ */
  ret += (200 - target_distance_delta) * 3;

  return (ret);
}

static int calc_priority_back_attack(int future, int future_step,
				     int moved, int moved_distance,
				     int target_distance_delta, int angle)
{
  int ret = 0;

  /* ʤȤäƤʤΤͥ */
  ret += ((future < 0) ? 200 : future) * 100;

  ret += moved;
  ret += moved_distance;

  /* ʤ٤ʤΥƤ */
  ret += (200 - target_distance_delta) * 3;

  /* åȤΤۤ */
  ret += (200 - calc_abs(angle)) * 4;

  return (ret);
}

static int calc_priority_random_walk(int future, int future_step,
				     int moved, int moved_distance,
				     int target_distance_delta, int angle)
{
  int ret = 0;

  /* ʤȤäƤʤΤͥ */
  ret += ((future < 0) ? 200 : future) * 100;

  ret += moved * 3;
  ret += moved_distance * 3;

  return (ret);
}

static int select_next_state_danger(int best_future,
				    int best_future_step,
				    int best_priority,
				    int best_moved,
				    int best_moved_distance,
				    int best_target_distance,
				    int best_target_distance_delta,
				    int best_target_angle,
				    Tank target,
				    int * timer)
{
  if ((best_future_step > -1) && (best_future_step < 5)) {
    return (ENEMY_STATE_DANGER);
  }
  return (ENEMY_STATE_ESCAPE);
}

static int select_next_state_escape(int best_future,
				    int best_future_step,
				    int best_priority,
				    int best_moved,
				    int best_moved_distance,
				    int best_target_distance,
				    int best_target_distance_delta,
				    int best_target_angle,
				    Tank target,
				    int * timer)
{
  if ((best_future_step > -1) && (best_future_step < 5)) {
    *timer = 0;
    return (ENEMY_STATE_DANGER);
  }
  if (best_future_step > -1)
    return (ENEMY_STATE_ESCAPE);
  if (GetRandomValue(4)) {
    return (ENEMY_STATE_CHASE);
  }
  return (ENEMY_STATE_BACK_ATTACK);
}

static int select_next_state_near(int best_future,
				  int best_future_step,
				  int best_priority,
				  int best_moved,
				  int best_moved_distance,
				  int best_target_distance,
				  int best_target_distance_delta,
				  int best_target_angle,
				  Tank target,
				  int * timer)
{
  if (best_future_step > -1) {
    *timer = 0;
    return (ENEMY_STATE_ESCAPE);
  }
  if (target == NULL)
    return (ENEMY_STATE_RANDOM_WALK);
  if (best_moved_distance < 20) {
    switch (GetRandomValue(4)) {
    case 0:
    case 1:
    case 2:
      return (ENEMY_STATE_BACK_ATTACK);
    case 3:
      return (ENEMY_STATE_BACK);
    }
  }
  if (best_target_distance < 3) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAREST);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  if (best_target_distance < 6) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAR);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  return (ENEMY_STATE_CHASE);
}

static int select_next_state_nearest(int best_future,
				     int best_future_step,
				     int best_priority,
				     int best_moved,
				     int best_moved_distance,
				     int best_target_distance,
				     int best_target_distance_delta,
				     int best_target_angle,
				     Tank target,
				     int * timer)
{
  if (best_future_step > -1) {
    *timer = 0;
    return (ENEMY_STATE_ESCAPE);
  }
  if (target == NULL)
    return (ENEMY_STATE_RANDOM_WALK);
  if (best_moved_distance < 20) {
    switch (GetRandomValue(4)) {
    case 0:
    case 1:
    case 2:
      return (ENEMY_STATE_BACK_ATTACK);
    case 3:
      return (ENEMY_STATE_BACK);
    }
  }
  if (best_target_distance < 3) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAREST);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  if (best_target_distance < 6) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAR);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  return (ENEMY_STATE_CHASE);
}

static int select_next_state_chase(int best_future,
				   int best_future_step,
				   int best_priority,
				   int best_moved,
				   int best_moved_distance,
				   int best_target_distance,
				   int best_target_distance_delta,
				   int best_target_angle,
				   Tank target,
				   int * timer)
{
  if (best_future_step > -1) {
    *timer = 0;
    return (ENEMY_STATE_ESCAPE);
  }
  if (target == NULL)
    return (ENEMY_STATE_RANDOM_WALK);
  if (best_moved_distance < 20) {
    switch (GetRandomValue(7)) {
    case 0:
    case 1:
      return (ENEMY_STATE_CHASE);
    case 2:
    case 3:
      return (ENEMY_STATE_CHASE_LEFT);
    case 4:
    case 5:
      return (ENEMY_STATE_CHASE_RIGHT);
    case 6:
      return (ENEMY_STATE_RANDOM_WALK);
    }
  }
  if (best_target_distance < 3) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAREST);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  if (best_target_distance < 6) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAR);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  return (ENEMY_STATE_CHASE);
}

static int select_next_state_chase_left(int best_future,
					int best_future_step,
					int best_priority,
					int best_moved,
					int best_moved_distance,
					int best_target_distance,
					int best_target_distance_delta,
					int best_target_angle,
					Tank target,
					int * timer)
{
  if (best_future_step > -1) {
    *timer = 0;
    return (ENEMY_STATE_ESCAPE);
  }
  if (target == NULL)
    return (ENEMY_STATE_RANDOM_WALK);
  if (best_moved_distance < 20) {
    switch (GetRandomValue(4)) {
    case 0:
      return (ENEMY_STATE_CHASE);
    case 1:
      return (ENEMY_STATE_CHASE_RIGHT);
    case 2:
      return (ENEMY_STATE_BACK);
    case 3:
      return (ENEMY_STATE_RANDOM_WALK);
    }
  }
  if (best_target_distance < 3) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAREST);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  if (best_target_distance < 6) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAR);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  if (GetRandomValue(4) == 0)
    return (ENEMY_STATE_CHASE);
  return (ENEMY_STATE_CHASE_LEFT);
}

static int select_next_state_chase_right(int best_future,
					 int best_future_step,
					 int best_priority,
					 int best_moved,
					 int best_moved_distance,
					 int best_target_distance,
					 int best_target_distance_delta,
					 int best_target_angle,
					 Tank target,
					 int * timer)
{
  if (best_future_step > -1) {
    *timer = 0;
    return (ENEMY_STATE_ESCAPE);
  }
  if (target == NULL)
    return (ENEMY_STATE_RANDOM_WALK);
  if (best_moved_distance < 20) {
    switch (GetRandomValue(4)) {
    case 0:
      return (ENEMY_STATE_CHASE);
    case 1:
      return (ENEMY_STATE_CHASE_LEFT);
    case 2:
      return (ENEMY_STATE_BACK);
    case 3:
      return (ENEMY_STATE_RANDOM_WALK);
    }
  }
  if (best_target_distance < 3) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAREST);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  if (best_target_distance < 6) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAR);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  if (GetRandomValue(4) == 0)
    return (ENEMY_STATE_CHASE);
  return (ENEMY_STATE_CHASE_RIGHT);
}

static int select_next_state_chase_back(int best_future,
					int best_future_step,
					int best_priority,
					int best_moved,
					int best_moved_distance,
					int best_target_distance,
					int best_target_distance_delta,
					int best_target_angle,
					Tank target,
					int * timer)
{
  if (best_future_step > -1) {
    *timer = 0;
    return (ENEMY_STATE_ESCAPE);
  }
  if (target == NULL)
    return (ENEMY_STATE_RANDOM_WALK);
  if (best_moved_distance < 20) {
    switch (GetRandomValue(4)) {
    case 0:
      return (ENEMY_STATE_CHASE);
    case 1:
      return (ENEMY_STATE_CHASE_LEFT);
    case 2:
      return (ENEMY_STATE_CHASE_RIGHT);
    case 3:
      return (ENEMY_STATE_RANDOM_WALK);
    }
  }
  if (best_target_distance < 3) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAREST);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  if (best_target_distance < 6) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAR);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  if (GetRandomValue(4) == 0)
    return (ENEMY_STATE_CHASE);
  return (ENEMY_STATE_BACK);
}

static int select_next_state_chase_back_attack(int best_future,
					       int best_future_step,
					       int best_priority,
					       int best_moved,
					       int best_moved_distance,
					       int best_target_distance,
					       int best_target_distance_delta,
					       int best_target_angle,
					       Tank target,
					       int * timer)
{
  if (best_future_step > -1) {
    *timer = 0;
    return (ENEMY_STATE_ESCAPE);
  }
  if (target == NULL)
    return (ENEMY_STATE_RANDOM_WALK);
  if (best_moved_distance < 20) {
    switch (GetRandomValue(4)) {
    case 0:
      return (ENEMY_STATE_CHASE);
    case 1:
      return (ENEMY_STATE_CHASE_LEFT);
    case 2:
      return (ENEMY_STATE_CHASE_RIGHT);
    case 3:
      return (ENEMY_STATE_RANDOM_WALK);
    }
  }
  if (best_target_distance < 3) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAREST);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  if (best_target_distance < 6) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAR);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  if (GetRandomValue(4) == 0)
    return (ENEMY_STATE_CHASE);
  return (ENEMY_STATE_BACK_ATTACK);
}

static int select_next_state_chase_random_walk(int best_future,
					       int best_future_step,
					       int best_priority,
					       int best_moved,
					       int best_moved_distance,
					       int best_target_distance,
					       int best_target_distance_delta,
					       int best_target_angle,
					       Tank target,
					       int * timer)
{
  if (best_future_step > -1) {
    *timer = 0;
    return (ENEMY_STATE_ESCAPE);
  }
  if (target == NULL)
    return (ENEMY_STATE_RANDOM_WALK);
  if (best_moved_distance < 20) {
    switch (GetRandomValue(4)) {
    case 0:
      return (ENEMY_STATE_CHASE);
    case 1:
      return (ENEMY_STATE_CHASE_LEFT);
    case 2:
      return (ENEMY_STATE_CHASE_RIGHT);
    case 3:
      return (ENEMY_STATE_BACK);
    }
  }
  if (best_target_distance < 3) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAREST);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  if (best_target_distance < 6) {
    if (GetRandomValue(3))
      return (ENEMY_STATE_NEAR);
    else
      return (ENEMY_STATE_BACK_ATTACK);
  }
  if (GetRandomValue(4) == 0)
    return (ENEMY_STATE_CHASE);
  return (ENEMY_STATE_RANDOM_WALK);
}

struct {

  /* åȤѹΨ */
  int change_target_rate;

  /* Ǥ᤯ˤ򥿡åȤȤΨ */
  int select_nearest_rate;

  /* ưΥ롼ײ*/
  int select_action_loop_count;

  /* ưοåȤФ١ե饰 */
  int random_action_depth;
  int random_action_count;
  int random_action_angle;
  int random_action_timer;
  unsigned long int random_action_flags;

  /* ưȽǤϰ */
  int random_action_range;

  /* ưκŬٿη׻Ѥδؿ */
  int (*calc_priority)(int future, int best_future_step,
		       int moved, int moved_distance,
		       int target_distance_delta, int angle);
  int (*select_next_state)(int best_future, int best_future_step,
			   int best_priority, int best_moved,
			   int best_moved_distance,
			   int best_target_distance,
			   int best_target_distance_delta,
			   int best_target_angle,
			   Tank target,
			   int * timer);

  /* ʿ(ޤƯƤ)ɤ߿ */
  int everytime_calc_future_depth;
} action_params[ENEMY_STATE_NUM] = {
  { 
    10, 100, 
    15,
    1, 8, 0, 3,
    RANDOM_ACTION_USE_STAY,
    15,
    calc_priority_danger,
    select_next_state_danger,
    4,
  }, /* danger */
  { 
    10, 100, 
    10,
    2, 5, 0, 3,
    RANDOM_ACTION_NO_FLAG,
    15,
    calc_priority_escape,
    select_next_state_escape,
    5,
  }, /* escape */
  {
    20, 100, 
    7,
    3, 3, 0, 5,
    RANDOM_ACTION_TURN_TO_TARGET | RANDOM_ACTION_FORWARD_OR_BACKWARD,
    0,
    calc_priority_near,
    select_next_state_near,
    10,
  }, /* near */
  {
    20, 100, 
    7,
    3, 3, 0, 5,
    RANDOM_ACTION_TURN_TO_TARGET,
    0,
    calc_priority_nearest,
    select_next_state_nearest,
    10,
  }, /* nearest */
  {
    3,  90,
    7,
    3, 3, 0, 5,
    RANDOM_ACTION_TURN_TO_TARGET | RANDOM_ACTION_FORWARD_ONLY,
    0,
    calc_priority_chase,
    select_next_state_chase,
    10,
  }, /* chase */
  {
    3,  90,
    7,
    3, 3, 90, 5,
    RANDOM_ACTION_TURN_TO_TARGET | RANDOM_ACTION_FORWARD_ONLY,
    0,
    calc_priority_chase_left,
    select_next_state_chase_left,
    10,
  }, /* chase_left */
  {
    3,  90,
    7,
    3, 3, -90, 5,
    RANDOM_ACTION_TURN_TO_TARGET | RANDOM_ACTION_FORWARD_ONLY,
    0,
    calc_priority_chase_right,
    select_next_state_chase_right,
    10,
  }, /* chase_right */
  {
    3,  80,
    7,
    3, 3, 0, 5,
    RANDOM_ACTION_NO_FLAG,
    0,
    calc_priority_back,
    select_next_state_chase_back,
    10,
  }, /* back */
  {
    3,  80,
    7,
    3, 3, 0, 5,
    RANDOM_ACTION_TURN_TO_TARGET | RANDOM_ACTION_BACKWARD_ONLY,
    0,
    calc_priority_back_attack,
    select_next_state_chase_back_attack,
    10,
  }, /* back_attack */
  {
    1,  70,
    7,
    3, 3, 0, 5,
    RANDOM_ACTION_FORWARD_ONLY,
    0,
    calc_priority_random_walk,
    select_next_state_chase_random_walk,
    10,
  }, /* random_walk */
};

static int EnemyAction(Enemy enemy, Map map,
		       ObjList tank_list, ObjList canon_list,
		       ObjList missile_list, ObjList explosion_list,
		       int width, int height, int peeping)
{
  int future, best_future = 0;
  int best_actions_num = 0;
  unsigned long int best_actions[ENEMY_ACTIONS_BUFFER_SIZE];
  int i, j;
  int future_step, best_future_step = 0;
  int moved, best_moved = 0;
  int moved_distance, best_moved_distance = 0;
  int target_distance, best_target_distance = 0;
  int target_distance_delta, best_target_distance_delta = 0;
  int target_angle, best_target_angle = 0;
  int priority, best_priority;
  int next_state;

  if ((enemy->target == NULL) ||
      (TankGetState(enemy->target) != TANK_ACTIVE) ||
      (GetRandomValue(100) < action_params[enemy->state].change_target_rate)) {
    select_target(enemy, tank_list,
		  action_params[enemy->state].select_nearest_rate,
		  width, height, peeping);
  }

  best_priority = -1;
  best_moved_distance = 0;

  for (i = 0;
       i < action_params[enemy->state].select_action_loop_count;
       i++) {
    set_random_action(enemy,
		      action_params[enemy->state].random_action_depth,
		      action_params[enemy->state].random_action_count,
		      action_params[enemy->state].random_action_angle,
		      action_params[enemy->state].random_action_flags);

    future = calc_future(ESCAPE_FLAG,
			 enemy, map, tank_list, canon_list,
			 missile_list, explosion_list,
			 width, height, peeping,
			 action_params[enemy->state].random_action_range,
			 action_params[enemy->state].random_action_depth *
			 action_params[enemy->state].random_action_count,
			 &future_step,
			 &moved,
			 &moved_distance,
			 &target_distance,
			 &target_distance_delta,
			 &target_angle);

    priority =
      (*(action_params[enemy->state].calc_priority))
      (future, future_step, moved, moved_distance, target_distance_delta,
       target_angle);

    if (best_priority > priority) continue;

    best_future = future;
    best_future_step = future_step;
    best_priority = priority;
    best_moved = moved;
    best_moved_distance = moved_distance;
    best_target_distance = target_distance;
    best_target_distance_delta = target_distance_delta;
    best_target_angle = target_angle;

    best_actions_num = enemy->actions_num;
    for (j = 0; j < enemy->actions_num; j++) {
      best_actions[j] = enemy->actions[j];
    }
  }

  enemy->current_action = 0;
  enemy->actions_num = best_actions_num;
  for (j = 0; j < enemy->actions_num; j++) {
    enemy->actions[j] = best_actions[j];
  }

  if (action_params[enemy->state].random_action_timer)
    enemy->timer = action_params[enemy->state].random_action_timer;
  else
    enemy->timer = enemy->actions_num;

  /* ޤˤ٤Τ餮 */
  if ((enemy->timer > 2) && (GetRandomValue(4) == 0))
    (enemy->timer)--;

  next_state =
    (*(action_params[enemy->state].select_next_state))
    (best_future, best_future_step,
     best_priority, best_moved, best_moved_distance,
     best_target_distance, best_target_distance_delta, best_target_angle,
     enemy->target, &(enemy->timer));

  /* ưڡ̵ʤȤꤢäƤޤ˲롪 */
  if (best_moved_distance < 20)
    enemy->actions[enemy->current_action] |= TANK_FIRE_CANON;

  return (next_state);
}

void print_state(Enemy enemy)
{
  static char * state[ENEMY_STATE_NUM] = {
    "ENEMY_STATE_DANGER",
    "ENEMY_STATE_ESCAPE",
    "ENEMY_STATE_NEAR",
    "ENEMY_STATE_NEAREST",
    "ENEMY_STATE_CHASE",
    "ENEMY_STATE_CHASE_LEFT",
    "ENEMY_STATE_CHASE_RIGHT",
    "ENEMY_STATE_BACK",
    "ENEMY_STATE_BACK_ATTACK",
    "ENEMY_STATE_RANDOM_WALK"
  };
  if (enemy == NULL) {
    fprintf(stderr, "Enemy is NULL.\n");
    return;
  }
  if (enemy->state >= ENEMY_STATE_NUM) {
    fprintf(stderr, "Unknown state.\n");
    return;
  }
  fprintf(stderr, "%d, %s\n", enemy->state, state[enemy->state]);
  return;
}

unsigned long int EnemyAct(Enemy enemy, Map map_orig,
			   ObjList player_list, ObjList enemy_list,
			   ObjList tank_list, ObjList canon_list,
			   ObjList missile_list, ObjList explosion_list,
			   int width, int height, int peeping)
{
  Map map;
  unsigned long int action;
  int future;

  if (TankGetState(PlayerGetTank((Player)enemy)) != TANK_ACTIVE) {
    return (TANK_NO_ACTION);
  }

  map = MapCreateFromMap(map_orig);

  if (enemy->timer) {
    if (enemy->current_action >= enemy->actions_num) {
      enemy->timer = 0;
    } else {
      (enemy->timer)--;
    }
  }

  if (enemy->timer) {
    future = calc_future(ESCAPE_FLAG, enemy, map, tank_list, canon_list,
			 missile_list, explosion_list,
			 width, height, peeping,
			 0,
			 action_params[enemy->state].everytime_calc_future_depth,
			 NULL, NULL, NULL, NULL, NULL, NULL);
    if (future >= 0)
      enemy->timer = 0;
  }

  while (enemy->timer == 0) {
    enemy->state = EnemyAction(enemy, map, tank_list, canon_list, missile_list,
			       explosion_list, width, height, peeping);
#if 0
    print_state(enemy);
#endif
  }

  shot_canon_and_missile(enemy, map, tank_list, canon_list,
			 missile_list, explosion_list, width, height, peeping);

  MapDestroy(map);
  if (enemy->actions_num > enemy->current_action) {
    action = enemy->actions[enemy->current_action];
    (enemy->current_action)++;
  } else {
    action = TANK_NO_ACTION;
  }

  return (action);
}

int EnemyInitActions(Enemy enemy)
{
  enemy->actions_num = 0;
  enemy->current_action = 0;
  return (0);
}

/*===========================================================================*/
/* ֥ȤȺ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

Enemy EnemyCreate(char * name, int state, int type)
{
  Enemy enemy;

  enemy = (Enemy)malloc(sizeof(_Enemy));
  if (enemy == NULL) ErrorExit("Cannot create enemy.");

  PlayerInitialize((Player)enemy, name, NULL, state);

  if (type == ENEMY_TYPE_RANDOM)
    enemy->type = ENEMY_TYPE_START + GetRandomValue(ENEMY_TYPE_NUMBER);
  else
    enemy->type = type;

  enemy->target = NULL;

  enemy->state = ENEMY_STATE_CHASE;
  enemy->timer = 0;

  enemy->actions_num = 0;
  enemy->current_action = 0;

  return (enemy);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

Enemy EnemyDestroy(Enemy enemy)
{
  if (enemy == NULL) return (NULL);
  PlayerDone((Player)enemy);
  free(enemy);
  return (NULL);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
