/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000-2003 Sakai Hiroaki.                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* Missile 饹(ߥѤΥ饹)Υ饤֥                  */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "MissileP.h"

/*****************************************************************************/
/* ؿ                                                                */
/*****************************************************************************/

/*===========================================================================*/
/* ѿμ                                                          */
/*===========================================================================*/

double MissileGetX( Missile missile) { return (missile->x ); }
double MissileGetY( Missile missile) { return (missile->y ); }
double MissileGetVX(Missile missile) { return (missile->vx); }
double MissileGetVY(Missile missile) { return (missile->vy); }
int MissileGetState(Missile missile)         { return (missile->state); }
int MissileSetState(Missile missile, int s)  { return (missile->state = s); }
Tank MissileGetTank(Missile missile)         { return (missile->tank); }
Tank MissileSetTank(Missile missile, Tank t) { return (missile->tank = t); }
Tank MissileGetTarget(Missile missile)       { return (missile->target); }
Tank MissileSetTarget(Missile missile, Tank tank)
{
  missile->tank = tank;
  if (missile->tank) TankIncMissile(missile->tank);
  return (missile->tank);
}

Tank MissileChangeTank(Missile missile, Tank tank)
{
  if (missile->tank) TankDecMissile(missile->tank);
  return (MissileSetTank(missile, tank));
}

/*===========================================================================*/
/* ȯȽ                                                                */
/*===========================================================================*/

static int MissileCheckExplosion(Missile missile, Map map, ObjList tank_list,
				 ObjList explosion_list)
{
  Tank tank;
  Explosion explosion;
  double x, y, d, size;
  ObjListData current;

  if (MapGetState(map,
		  ConvertXFromPositionToMap(missile->x),
		  ConvertYFromPositionToMap(missile->y))) {
#if 0 /* ޥåץǡιϡExplosion ֥ȤԤ */
    MapChangeState(map,
		   ConvertXFromPositionToMap(missile->x),
		   ConvertYFromPositionToMap(missile->y), 0);
#endif
    missile->state = MISSILE_EXPLOSION;
    return (missile->state);
  }

  size = (double)EXPLOSION_SIZE / CHARACTER_SIZE * 0.5 * 1.2;

  for (current = ObjList_GetStart(tank_list);
       !ObjList_IsEndEdge(tank_list, current);
       current = ObjListData_GetNext(current)) {
    tank = (Tank)ObjListData_GetObj(current);
    x = missile->x - TankGetX(tank);
    y = missile->y - TankGetY(tank);
    d = x * x + y * y;
    if (d < size * size) {
      missile->state = MISSILE_EXPLOSION;
      return (missile->state);
    }
  }

  for (current = ObjList_GetStart(explosion_list);
       !ObjList_IsEndEdge(explosion_list, current);
       current = ObjListData_GetNext(current)) {
    explosion = (Explosion)ObjListData_GetObj(current);
    if (ExplosionCheck(explosion, missile->x, missile->y)) {
      missile->state = MISSILE_EXPLOSION;
      return (missile->state);
    }
  }

  return (missile->state);
}

/*===========================================================================*/
/* åȤθ                                                          */
/*===========================================================================*/

int MissileSearchTarget(Missile missile, ObjList tank_list)
{
  double dx, dy, d = -1.0;
  Tank tank, target;
  ObjListData current;

  /* ֶ᤯ˤ륿åȤ򸡺 */
  target = NULL;
  for (current = ObjList_GetStart(tank_list);
       !ObjList_IsEndEdge(tank_list, current);
       current = ObjListData_GetNext(current)) {
    tank = (Tank)ObjListData_GetObj(current);
    if (tank == missile->tank) continue;
    if (TankGetState(tank) == TANK_DESTROY) continue;
    dx = TankGetX(tank) - missile->x;
    dy = TankGetY(tank) - missile->y;
    if ((d < 0) || (dx * dx + dy * dy < d)) {
      d = dx * dx + dy * dy;
      target = tank;
    }
  }
  missile->target = target;
  return (0);
}

/*===========================================================================*/
/* ֥ȤȺ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

static Missile MissileAllocate(double x, double y, double vx, double vy,
			       Tank tank, int state, Tank target)
{
  Missile missile;

  missile = (Missile)malloc(sizeof(_Missile));
  if (missile == NULL) ErrorExit("Cannot create missile.");

  missile->x = x + vx * MISSILE_OFFSET;
  missile->y = y + vy * MISSILE_OFFSET;
  missile->vx = vx;
  missile->vy = vy;
  missile->tank = tank;
  missile->state = state;
  missile->target = target;

  return (missile);
}

Missile MissileCreate(double x, double y, double vx, double vy, Tank tank,
		      int state, Tank target)
{
  Missile missile;

  missile = MissileAllocate(x, y, vx, vy, tank, state, target);
  if (missile == NULL) ErrorExit("Cannot create missile.");
  if (missile->tank) TankIncMissile(missile->tank);

  return (missile);
}

Missile MissileCreateFromMissile(Missile old_missile)
{
  Missile missile;

  missile = MissileAllocate(old_missile->x, old_missile->y,
			    old_missile->vx, old_missile->vy,
			    old_missile->tank, old_missile->state,
			    old_missile->target);
  if (missile == NULL) ErrorExit("Cannot create missile.");

  /*
   * ԡƺˤϡ󥯥ȤԤʤ
   * ( MissileSetTank() ꤹ٤)
   *  destroy ˤϥǥȤԤΤա
   * (Τޤ destroy ȡ礬)
   */
  /* if (missile->tank) TankIncMissile(missile->tank); */

  return (missile);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

Missile MissileDestroy(Missile missile)
{
  if (missile == NULL) return (NULL);

  if (missile->tank) TankDecMissile(missile->tank);
  free(missile);

  return (NULL);
}

/*===========================================================================*/
/* ư                                                                  */
/*===========================================================================*/

static int MissileMove(Missile missile, Map map)
{
  double dx, dy, d, vd;

  if (missile->target) {
    dx = TankGetX(missile->target) - missile->x;
    dy = TankGetY(missile->target) - missile->y;
#if 0 /* libm ϻѤʤ */
    d = sqrt(dx*dx + dy*dy);
#else
    d = easy_sqrt(dx*dx + dy*dy);
#endif
    if (d > 0.01) {
      missile->vx += dx * MISSILE_QUICK / d;
      missile->vy += dy * MISSILE_QUICK / d;
#if 0 /* libm ϻѤʤ */
      vd = sqrt(missile->vx * missile->vx + missile->vy * missile->vy);
#else
      vd = easy_sqrt(missile->vx * missile->vx + missile->vy * missile->vy);
#endif
      if (vd > MISSILE_SPEED) {
	missile->vx *= MISSILE_SPEED / vd;
	missile->vy *= MISSILE_SPEED / vd;
      }
    }
  }

  missile->x += missile->vx;
  missile->y += missile->vy;

  return (0);
}

int MissileAct(Missile missile, Map map, ObjList tank_list,
	       ObjList explosion_list)
{
  switch (missile->state) {
  case MISSILE_WAIT :
    break;
  case MISSILE_ACTIVE :
    MissileMove(missile, map);
    MissileCheckExplosion(missile, map, tank_list, explosion_list);
    break;
  case MISSILE_EXPLOSION :
    break;
  case MISSILE_DESTROY :
    break;
  default :
    break;
  }
  return (missile->state);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
