/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000-2003 Sakai Hiroaki.                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* Player 饹(ץ쥤䡼Υ饹)Υ饤֥                       */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "PlayerP.h"

char * player_names[] = {
  "Bolken", "WildDog", "Hornet", "Airi", "Akana",
  "GreenTea", "Dindon", "Ripple", "RingOnTheWater", "Blaky",
  "HalfPot", "SmartBomb", "Flogger", "DirtyHorse", "HungryStomach",
  "SlamPoint", "RattleSnake", "GrindBoy", "SqueakVoice", "JaggerEdge",
  "Belch", "WoodenDoll", "NimbleBoy", "PlayingHookey", "DrippingWet",
  "SillyBoy", "VeryBottom", "RapidlyLike", "Nawo", "ScrewDriver",
  "NezumiKozou", "Nababa", "Yuuki", "NipponoFace", "WhoAmI?",
  "FuurinKazan", "Polpfun", "LZ", "I.L.J.", "Folcman",
  "Goemon", "Jorjor", "BerkenIII", "Kame", "Sushi",
  "JarringSound", "Jumble", "MixJuice", "CupCake", "CottageCheese",
  "Gingin", "Kennel", "CoffeeBean", "JunkFood", "HighwayPatrol"
};

#define PLAYER_NAMES_NUMBER ((int)(sizeof(player_names)/sizeof(player_names[0])))

/*****************************************************************************/
/* ؿ                                                                */
/*****************************************************************************/

/*===========================================================================*/
/* ѿμ                                                    */
/*===========================================================================*/

int PlayerGetState(Player p)            { return (p->state        ); }
int PlayerSetState(Player p, int state) { return (p->state = state); }
Path PlayerGetPath(Player p)            { return (p->path         ); }
Path PlayerSetPath(Player p, Path path) { return (p->path  = path ); }
Tank PlayerGetTank(Player p)            { return (p->tank         ); }
Tank PlayerSetTank(Player p, Tank tank) { return (p->tank  = tank ); }
char * PlayerGetName(Player p) { return (p->name); }
int PlayerGetColorNumber(Player p) { return (p->color_number); }
unsigned long int PlayerGetJoinTime(Player p) { return (p->join_time); }

int PlayerGetWin( Player p) {return (   p->win  );}
int PlayerIncWin( Player p) {return (++(p->win ));}
int PlayerGetLose(Player p) {return (   p->lose );}
int PlayerIncLose(Player p) {return (++(p->lose));}

/*===========================================================================*/
/* ֥ȤȺ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

int PlayerInitialize(Player player, char * name, Path path, int state)
{
  static int color_number = 1;

  if (player == NULL) return (0);

  player->win  = 0;
  player->lose = 0;
  if (name) {
    if (!strcmp(name, DEFAULT_PLAYER_NAME))
      player->name = strdup(player_names[GetRandomValue(PLAYER_NAMES_NUMBER)]);
    else
      player->name = strdup(name);
  } else{
    player->name = strdup("Who am I?");
  }

  /* ץ쥤䡼̾Ĺ */
  if (strlen(player->name) > PLAYER_NAME_MAX_LENGTH) {
    (player->name)[PLAYER_NAME_MAX_LENGTH] = '\0';
  }

  player->path = path;
  player->tank = NULL;
  player->state = state;
  player->color_number = color_number++;
  player->join_time = GetNowTime(); /* ߻ǼƤ */

  return (0);
}

Player PlayerCreate(char * name, Path path, int state)
{
  Player player;

  player = (Player)malloc(sizeof(_Player));
  if (player == NULL) ErrorExit("Cannot create player.");

  PlayerInitialize(player, name, path, state);

  return (player);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

int PlayerDone(Player player)
{
  if (player == NULL) return (0);
  if (player->name) free(player->name);
  if (player->path) PathDestroy(player->path);
  if (player->tank) player->tank = NULL; /* ƤϤʤ */
  return (0);
}

Player PlayerDestroy(Player player)
{
  if (player == NULL) return (NULL);
  PlayerDone(player);
  free(player);
  return (NULL);
}

/*===========================================================================*/
/* ǡ                                                            */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ǡɤ߹                                                      */
/*---------------------------------------------------------------------------*/

char * PlayerReadData(Player player, char * p)
{
  char command[30];
  char name[50];

  if (!player) return (NULL);
  if (!p     ) return (NULL);

  sscanf(p, "%s\t%d %d %d %s %d\n",
	 command, &(player->state), &(player->win), &(player->lose),
	 name, &(player->color_number));

  if (player->name) free(player->name);
  player->name = strdup(name);

  while (*(p++) != '\n') { /* None */ }

  return (p);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
