/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000-2003 Sakai Hiroaki.                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* Stage 饹(ѤΥ饹)Υ饤֥                        */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "StageP.h"

#ifdef USE_ALARM_SIGNAL
#include <signal.h>
#endif

/*****************************************************************************/
/* ؿ                                                                */
/*****************************************************************************/

/*===========================================================================*/
/* ѿμ                                                    */
/*===========================================================================*/

Map StageGetMap(Stage s) {return (s->map);}
ObjList StageGetPlayerList(   Stage s) {return (s->player_list   );}
ObjList StageGetTankList(     Stage s) {return (s->tank_list     );}
ObjList StageGetCanonList(    Stage s) {return (s->canon_list    );}
ObjList StageGetMissileList(  Stage s) {return (s->missile_list  );}
ObjList StageGetExplosionList(Stage s) {return (s->explosion_list);}

/*===========================================================================*/
/* ֥ȤȺ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

int StageCreateTankForPlayer(Stage stage, Player player)
{
  Tank tank;
  int x, y, tx, ty, f;
  ObjListData current;
  double rx, ry, dx, dy, r;
  Tank other;

  r = ((double)TANK_SIZE / CHARACTER_SIZE) * 2.0;

  /* ᤯̵̤֤ɤå */
  do {
    rx = GetRandomValue(stage->map_width ) + 0.5;
    ry = GetRandomValue(stage->map_height) + 0.5;
    f = 0;
    for (current = ObjList_GetStart(stage->tank_list);
	 !ObjList_IsEndEdge(stage->tank_list, current);
	 current = ObjListData_GetNext(current)) {
      other = (Tank)ObjListData_GetObj(current);
      dx = TankGetX(other) - rx;
      dy = TankGetY(other) - ry;
      if (dx * dx + dy * dy <= r * r) {
	f = 1;
	break;
      }
    }
  } while (f);

  tank = TankCreate(rx, ry, GetRandomValue(TANK_ANGLE_NUMBER),
		    player, TANK_ACTIVE, TANK_NO_ACTION, stage->backward);
  if (tank == NULL) ErrorExit("Cannot create tank.");
  PlayerSetTank(player, tank);
  ObjList_InsertObjToEnd(stage->tank_list, tank, (ObjDestructor)TankDestroy);

  /* ֤μ򥯥ꥢ */
  for (x = -1; x <= 1; x++) {
    for (y = -1; y <= 1; y++) {
      tx = ConvertXFromPositionToMap(TankGetX(tank)) + x;
      ty = ConvertYFromPositionToMap(TankGetY(tank)) + y;
      if (MapIsInside(stage->map, tx, ty))
	MapChangeState(stage->map, tx, ty, 0);
    }
  }
  return (0);
}

Stage StageCreate(int stage_number, ObjList player_list, ObjList enemy_list,
		  int map_width, int map_height, int width, int height,
		  int peeping, int backward,
		  int interval_time, int timeout_time, int max_players)
{
  Stage stage;
  Player player;
  Enemy enemy;
  ObjListData current;

  stage = (Stage)malloc(sizeof(_Stage));
  if (stage == NULL) return (NULL);

  stage->map_width   = map_width;
  stage->map_height  = map_height;
  stage->width       = width;
  stage->height      = height;
  stage->peeping     = peeping;
  stage->backward    = backward;
  stage->interval    = interval_time;
  stage->timeout     = timeout_time;
  stage->max_players = max_players;

  stage->map = MapCreate(stage->map_width, stage->map_height);
  if (stage->map == NULL) ErrorExit("Cannot create stage->map.");
  MapMake(stage->map, stage_number);

  stage->stage_number   = stage_number;
  stage->player_list    = player_list;
  stage->enemy_list     = enemy_list;
  stage->tank_list      = ObjList_Create();
  stage->canon_list     = ObjList_Create();
  stage->missile_list   = ObjList_Create();
  stage->explosion_list = ObjList_Create();

  for (current = ObjList_GetStart(player_list);
       !ObjList_IsEndEdge(player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    if (PlayerGetState(player) != PLAYER_ACTIVE) continue;
    StageCreateTankForPlayer(stage, player);
  }

  for (current = ObjList_GetStart(enemy_list);
       !ObjList_IsEndEdge(enemy_list, current);
       current = ObjListData_GetNext(current)) {
    enemy = (Enemy)ObjListData_GetObj(current);
    if (PlayerGetState((Player)enemy) != PLAYER_ACTIVE) continue;
    StageCreateTankForPlayer(stage, (Player)enemy);
  }

  return (stage);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

Stage StageDestroy(Stage stage)
{
  ObjListData current;
  Player player;

  if (stage == NULL) return (NULL);

  if (stage->map) MapDestroy(stage->map);

  for (current = ObjList_GetStart(stage->player_list);
       !ObjList_IsEndEdge(stage->player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    PlayerSetTank(player, NULL);
  }
  if (stage->tank_list)      ObjList_Destroy(stage->tank_list);
  if (stage->canon_list)     ObjList_Destroy(stage->canon_list);
  if (stage->missile_list)   ObjList_Destroy(stage->missile_list);
  if (stage->explosion_list) ObjList_Destroy(stage->explosion_list);

  free(stage);

  return (NULL);
}

/*---------------------------------------------------------------------------*/
/* Stage is over?                                                            */
/*---------------------------------------------------------------------------*/

static int StageIsOver(Stage stage, int lose)
{
  int n;

  n = ObjList_GetLength(stage->tank_list);

  switch (n) {
  case 0 :
    usleep(100000);
    return (0);
  case 1 :
    return (lose >= 1);
  case 2 :
    return (lose >= 1);
  case 3 :
    return (lose >= 2);
  case 4 :
    return (lose >= 2);
  case 5 :
    return (lose >= 3);
  default :
    return (lose >= n * 2 / 3);
  }
  return (0);
}

/*---------------------------------------------------------------------------*/
/* ƥȤɽ                                                            */
/*---------------------------------------------------------------------------*/

static int StageSendPlayerData(ObjList player_list, Path path)
{
  Player player;
  ObjListData current;
  char buffer[PLAYER_NAME_MAX_LENGTH + 100];
  for (current = ObjList_GetStart(player_list);
       !ObjList_IsEndEdge(player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    if (PlayerGetState(player) != PLAYER_ACTIVE) continue;
    if (strlen(PlayerGetName(player)) > PLAYER_NAME_MAX_LENGTH) {
      ErrorExit("Player\'s name is too long.");
    }
    sprintf(buffer, "   %s\n   Win:%d Lose:%d\n\n",
	    PlayerGetName(player),
	    PlayerGetWin(player),
	    PlayerGetLose(player));
    PathSendShortStringData(path, TEXT_DRAW, PlayerGetColorNumber(player),
			    buffer);
  }
  return (0);
}

static int StageSendTextData(Stage stage, Path path)
{
  char buffer[PLAYER_NAME_MAX_LENGTH + 100];

  PathSendNoneData(path, TEXT_CLEAR);
  PathSendShortStringData(path, TEXT_DRAW, 0,
			  "      N.A.D.A.R.\n  Copyright (c)\n");
  PathSendShortStringData(path, TEXT_DRAW, 0,
			  "           2000-2003\n    Sakai Hiroaki.\n\n");

  sprintf(buffer, "   Stage : %d\n\n", stage->stage_number + 1);
  PathSendShortStringData(path, TEXT_DRAW, 0, buffer);

  StageSendPlayerData(stage->player_list, path);
  StageSendPlayerData(stage->enemy_list, path);

  return (0);
}

/*===========================================================================*/
/* ơγ(ФΥᥤ롼)                                      */
/*===========================================================================*/

/* ȯν */
int StageProcExplosion(Map map, ObjList explosion_list)
{
  ObjListData current;
  Explosion explosion;
  int state;
  for (current = ObjList_GetStart(explosion_list);
       !ObjList_IsEndEdge(explosion_list, current);
       current = ObjListData_GetNext(current)) {
    explosion = (Explosion)ObjListData_GetObj(current);
    state = ExplosionAct(explosion, map);
    switch (state) {
    case EXPLOSION_DESTROY :
      current = ObjList_DeleteObjToPrev(explosion_list, current);
      break;
    default :
      break;
    }
  }
  return (0);
}

/* ˤƤν */
int StageProcCanon(Map map, ObjList tank_list, ObjList canon_list,
		   ObjList explosion_list)
{
  ObjListData current;
  Canon canon;
  Explosion explosion;
  int state;
  for (current = ObjList_GetStart(canon_list);
       !ObjList_IsEndEdge(canon_list, current);
       current = ObjListData_GetNext(current)) {
    canon = (Canon)ObjListData_GetObj(current);
    state = CanonAct(canon, map, tank_list, explosion_list);
    switch (state) {
    case CANON_EXPLOSION :
      explosion = ExplosionCreate(CanonGetX(canon), CanonGetY(canon),
				  CanonGetTank(canon), EXPLOSION_ACTIVE);
      ObjList_InsertObjToEnd(explosion_list, explosion,
			     (ObjDestructor)ExplosionDestroy);
      CanonSetState(canon, CANON_DESTROY);
      /* Fall through to destroy the object */
    case CANON_DESTROY :
      current = ObjList_DeleteObjToPrev(canon_list, current);
      break;
    default :
      break;
    }
  }
  return (0);
}

/* ߥν */
int StageProcMissile(Map map, ObjList tank_list, ObjList missile_list,
		     ObjList explosion_list)
{
  ObjListData current;
  Missile missile;
  Explosion explosion;
  int state;
  for (current = ObjList_GetStart(missile_list);
       !ObjList_IsEndEdge(missile_list, current);
       current = ObjListData_GetNext(current)) {
    missile = (Missile)ObjListData_GetObj(current);
    state = MissileAct(missile, map, tank_list, explosion_list);
    switch (state) {
    case MISSILE_EXPLOSION :
      explosion = ExplosionCreate(MissileGetX(missile),
				  MissileGetY(missile),
				  MissileGetTank(missile),
				  EXPLOSION_ACTIVE);
      ObjList_InsertObjToEnd(explosion_list, explosion,
			     (ObjDestructor)ExplosionDestroy);
      MissileSetState(missile, MISSILE_DESTROY);
      /* Fall through to destroy the object */
    case MISSILE_DESTROY :
      current = ObjList_DeleteObjToPrev(missile_list, current);
      break;
    default :
      break;
    }
  }
  return (0);
}

/* ֤ν */
int StageProcTank(Map map, ObjList tank_list, ObjList canon_list,
		  ObjList missile_list, ObjList explosion_list)
{
  ObjListData current;
  Tank tank;
  Explosion explosion;
  int state;
  for (current = ObjList_GetStart(tank_list);
       !ObjList_IsEndEdge(tank_list, current);
       current = ObjListData_GetNext(current)) {
    tank = (Tank)ObjListData_GetObj(current);
    state = TankAct(tank, map, tank_list, canon_list, missile_list,
		    explosion_list);
    switch (state) {
    case TANK_EXPLOSION :
      explosion = ExplosionCreate(TankGetX(tank), TankGetY(tank),
				  tank, EXPLOSION_ACTIVE);
      ObjList_InsertObjToEnd(explosion_list, explosion,
			     (ObjDestructor)ExplosionDestroy);
      TankSetState(tank, TANK_EXPLOSING);
      TankSetDestroyTimer(tank, TANK_DESTROY_INTERVAL);
      break;
    case TANK_EXPLOSING :
      TankDecDestroyTimer(tank);
      if (TankGetDestroyTimer(tank) <= 0) {
	TankSetState(tank, TANK_DESTROY);
      }
      break;
    case TANK_DESTROY :
      /* current = ObjList_DeleteObjToPrev(stage->tank_list, current); */
      break;
    default :
      break;
    }
  }
  return (0);
}

#ifdef USE_ALARM_SIGNAL
static void alarm_hundler()
{
  /* none */
}
#endif

#define BUFFER_SIZE (PLAYER_NAME_MAX_LENGTH + 32)

int StageStartAsServer(Stage stage, ObjList player_list, Path request)
{
  unsigned short int command;
  unsigned short int length;
  int ret;
  Path path;
  Player player;
  Player player2;
  Enemy enemy;
  Tank tank;
  Canon canon;
  Missile missile;
  Explosion explosion;
  int lose;
  ObjListData current, current2;
  unsigned long int l1;
  unsigned long int now_time;
  char player_name[BUFFER_SIZE];
  int dont_move;
  long int dont_move_timer_start;
  long int dont_move_timer;
#ifdef USE_ALARM_SIGNAL
  struct itimerval itval;
  Player * fd2player;
  int player_num;
  int fd;
  int max_fd = -1;
  fd_set readfds;
  fd_set fds;
  struct timeval timeout;
  struct timeval * timeout_p;
#else
  long int interval_timer;
  int f0, f1, f2;
  long int timeout_timer;
#endif

  /*
   * ХåեΥꥢ
   * Хåե򥯥ꥢƤʤȡơ׵᤬ĤäƤޤ
   * Ѥ߽ŤʤäƤäƤޤǽ롥
   * (SEND_KEY Ф NOW_KEY Υ꥿󤬤ޤäƤޤ̥ꥢ
   *  ӤȿߤʤäƤޤäƤ)
   */
  for (current = ObjList_GetStart(player_list);
       !ObjList_IsEndEdge(player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    if (PlayerGetState(player) != PLAYER_ACTIVE) continue;

    while (PathCheckData(PlayerGetPath(player), &command, &length)) {

      switch (command) {

      case DELETE_ME :

	PathRecvNoneData(PlayerGetPath(player), &command);
	PathSendNoneData(PlayerGetPath(player), DELETE_YOU);
	PathFlushSendBuffer(PlayerGetPath(player));
	if (TankGetState(PlayerGetTank(player)) == TANK_ACTIVE)
	  TankSetState(PlayerGetTank(player), TANK_EXPLOSION);
	PlayerSetState(player, PLAYER_DESTROY);
	break;

      default :
	/*
	 * ХåեˤޤäƤ׵򥯥ꥢ롥
	 * (ơ׵᤬ĤäƤޤɤɤѤ߽ŤʤäƤȤ)
	 */
	PathDiscardData(PlayerGetPath(player), &command);
	break;
      }
    }
  }

  for (current = ObjList_GetStart(player_list);
       !ObjList_IsEndEdge(player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    if (PlayerGetState(player) != PLAYER_ACTIVE) continue;

    /* ǥץ쥤Υꥢ */
    PathSendNoneData(PlayerGetPath(player), DISPLAY_CLEAR);

    /* ɽƥȥǡ */
    StageSendTextData(stage, PlayerGetPath(player));
  }

  /* ޥåץǡ */
  MapSendChangesToAllPlayers(stage->map, stage->player_list);

  for (current = ObjList_GetStart(player_list);
       !ObjList_IsEndEdge(player_list, current);
       current = ObjListData_GetNext(current)) {
    player = (Player)ObjListData_GetObj(current);
    if (PlayerGetState(player) != PLAYER_ACTIVE) continue;
#if 0 /* ̺̿︺ΤᡤSEND_KEY  XSync() ԤREAD_KEY פˤ */
    PathSendNoneData(PlayerGetPath(player), READ_KEY);
#endif
    PathSendNoneData(PlayerGetPath(player), SEND_KEY);

    /* Хåեեå夷ºݤ˥ǡ */
    PathFlushSendBuffer(PlayerGetPath(player));
  }

  /*
   * Ũν
   */
  for (current = ObjList_GetStart(stage->enemy_list);
       !ObjList_IsEndEdge(stage->enemy_list, current);
       current = ObjListData_GetNext(current)) {
    enemy = (Enemy)ObjListData_GetObj(current);
    EnemyInitActions(enemy);
  }

  /* ޤν */
  if (stage->interval) {
#ifdef USE_ALARM_SIGNAL
    itval.it_interval.tv_sec  =  stage->interval / 1000;
    itval.it_interval.tv_usec = (stage->interval % 1000) * 1000;
    itval.it_value.tv_sec     =  stage->interval / 1000;
    itval.it_value.tv_usec    = (stage->interval % 1000) * 1000;
    setitimer(ITIMER_REAL, &itval, NULL);
    signal(SIGALRM, alarm_hundler);
#else
    stage->interval_timer = GetTimeForMilliSeconds();
#endif
  }
#if 0
  if (stage->timeout) stage->timeout_timer  = GetTimeForMilliSeconds();
#endif

  dont_move = 1;
  dont_move_timer_start = GetTimeForMilliSeconds();

  while (1) { /* ơΥᥤ롼 */

    /* ʥˤ뽪λ */
    if (IsTerminated()) {
      break;
    }

    /* 饤Ȥμդ */

    if (PathCheckRecvState(request)) {

      path = PathCreateToCommunicateWithClient(request,
					       SERVER_SEND_BUFFER_SIZE,
					       SERVER_RECV_BUFFER_SIZE);

      while (!PathCheckRecvState(path)) { usleep(10000); }
      ret = PathCheckData(path, &command, &length);

      if (ret && (command == JOIN_REQUEST)) {

	PathRecvStringData(path, &command, player_name, BUFFER_SIZE);

	if (ObjList_GetLength(player_list) >= stage->max_players) {
	  PathSendNoneData(path, MANY_PLAYERS);
	  PathFlushSendBuffer(path);
	  path = PathDestroy(path);
	} else {
	  player = PlayerCreate(player_name, path, PLAYER_WAIT);
	  ObjList_InsertObjToEnd(player_list, player,
				 (ObjDestructor)PlayerDestroy);
	  PathSendShortShortShortShortShortData(PlayerGetPath(player),
						JOIN_ACCEPT,
						stage->map_width,
						stage->map_height,
						stage->width,
						stage->height,
						stage->peeping);
	  PathFlushSendBuffer(PlayerGetPath(player));
	}
      } else {
        path = PathDestroy(path);
      }
    }

    /* ȯν */
    StageProcExplosion(stage->map, stage->explosion_list);

    /* ˤƤν */
    StageProcCanon(stage->map, stage->tank_list, stage->canon_list,
		   stage->explosion_list);

    /* ߥν */
    StageProcMissile(stage->map, stage->tank_list, stage->missile_list,
		     stage->explosion_list);

    /*
     * Ũν
     */
    for (current = ObjList_GetStart(stage->enemy_list);
	 !ObjList_IsEndEdge(stage->enemy_list, current);
	 current = ObjListData_GetNext(current)) {
      enemy = (Enemy)ObjListData_GetObj(current);
      if (PlayerGetState((Player)enemy) != PLAYER_ACTIVE) continue;
      if (dont_move)
	l1 = TANK_NO_ACTION;
      else
	l1 = EnemyAct(enemy, stage->map, 
		      stage->player_list, stage->enemy_list,
		      stage->tank_list, stage->canon_list,
		      stage->missile_list, stage->explosion_list,
		      stage->width, stage->height, stage->peeping);
      TankSetAction(PlayerGetTank((Player)enemy), l1);
    }

    /*
     * Wait ֤Υץ쥤䡼ν
     */
    for (current = ObjList_GetStart(stage->player_list);
	 !ObjList_IsEndEdge(stage->player_list, current);
	 current = ObjListData_GetNext(current)) {

      player = (Player)ObjListData_GetObj(current);
      if (PlayerGetState(player) != PLAYER_WAIT) continue;

      /*
       * Wait ֤ˤʤäƤץ쥤䡼ϾʤΤǡWait ֤ˤʤäƤ
       * ץ쥤䡼ߤĤäȤ GetNowTime() ǻĴ٤нʬ
       * 뤫顤GetNowTime() Ϥ֤(for()֤ɬפ̵)
       */
      now_time = GetNowTime();

      /* 120äĶƤADD_MEƤʤˤϡץ쥤䡼 */
      if (now_time - PlayerGetJoinTime(player) > 120) {
	PathSendNoneData(PlayerGetPath(player), TIMEOUT_YOU);
	PathFlushSendBuffer(PlayerGetPath(player));
	PlayerSetState(player, PLAYER_DESTROY);
	continue;
      }

      if (!PathCheckData(PlayerGetPath(player), &command, &length))
	continue;

      switch (command) {

      case ADD_ME :
	PathRecvNoneData(PlayerGetPath(player), &command);

	/*
	 * MapSendAllMap() ñ˥ޥåפ򤽤ΤޤʤΤǡ
	 * MapChange 塼ޤäƤȡ MapChange äȤ
	 * 򵯤ΤᡤMapSendAllMap ¹Ԥ
	 * MapSendChangesToAllPlayers() ԤMapChange 塼եå
	 * 롥 MapChange ǡϿץ쥤䡼ˤƤϤʤ
	 * ΤǡPLAYER_ACTIVE ˤ˹Ԥɬפ롥
	 * StageCreateTankForPlayer() ǥޥåפѹ뤳ȤΤǡ
	 * MapSendChangesToAllPlayers() ǸƤӽФɬפ롥
	 */
	StageCreateTankForPlayer(stage, player);
	MapSendChangesToAllPlayers(stage->map, stage->player_list);

	PlayerSetState(player, PLAYER_ACTIVE);

	PathSendNoneData(PlayerGetPath(player), DISPLAY_CLEAR);

	MapSendAllMap(stage->map, player);

	/* ɽƥȥǡץ쥤䡼 */
	for (current2 = ObjList_GetStart(stage->player_list);
	     !ObjList_IsEndEdge(stage->player_list, current2);
	     current2 = ObjListData_GetNext(current2)) {
	  player2 = (Player)ObjListData_GetObj(current2);
	  if (PlayerGetState(player2) != PLAYER_ACTIVE) continue;
	  StageSendTextData(stage, PlayerGetPath(player2));
	  PathFlushSendBuffer(PlayerGetPath(player2));
	}

#if 0 /* ̺̿︺ΤᡤSEND_KEY  XSync() ԤREAD_KEY פˤ */
	PathSendNoneData(PlayerGetPath(player), READ_KEY);
#endif
	PathSendNoneData(PlayerGetPath(player), SEND_KEY);

	/* Хåեեå夷ºݤ˥ǡ */
	PathFlushSendBuffer(PlayerGetPath(player));

	break;

      case DELETE_ME :

	PathRecvNoneData(PlayerGetPath(player), &command);
	PathSendNoneData(PlayerGetPath(player), DELETE_YOU);
	PathFlushSendBuffer(PlayerGetPath(player));
	PlayerSetState(player, PLAYER_DESTROY);
	break;

      default :
	ErrorExit("Invalid command.");
      }
    }

    /*
     *  Player ǡȤ롥ǡä Player ϡ
     * State Ū Done ˤƤ
     */
#ifdef USE_ALARM_SIGNAL
    /*
     * ץ쥤䡼滲äƤ礬Τǡ max_fd 
     * ׻malloc() 롥
     * (Ūˤ realloc() Ѥ褦˽ۤ)
     */
    max_fd = -1;
    for (current = ObjList_GetStart(player_list);
	 !ObjList_IsEndEdge(player_list, current);
	 current = ObjListData_GetNext(current)) {
      player = (Player)ObjListData_GetObj(current);
      if (PlayerGetState(player) != PLAYER_ACTIVE) continue;
      fd = PathGetFileDescriptor(PlayerGetPath(player));
      if (max_fd < fd) max_fd = fd;
    }
    if (max_fd > -1) {
      fd2player = malloc(sizeof(Player) * (max_fd + 1));
      for (fd = 0; fd <= max_fd; fd++) {
	fd2player[fd] = NULL;
      }
    } else {
      fd2player = NULL;
    }

    FD_ZERO(&readfds);
    player_num = 0;
    for (current = ObjList_GetStart(player_list);
	 !ObjList_IsEndEdge(player_list, current);
	 current = ObjListData_GetNext(current)) {
      player = (Player)ObjListData_GetObj(current);
      if (PlayerGetState(player) != PLAYER_ACTIVE) continue;
      fd = PathGetFileDescriptor(PlayerGetPath(player));
      fd2player[fd] = player;
      FD_SET(fd, &readfds);
      player_num++;
    }

    if (stage->timeout) {
      if (stage->interval) {
	stage->timeout_timer = stage->timeout + stage->interval;
      } else {
	stage->timeout_timer = stage->timeout;
      }
    }
    timeout_p = NULL;

    while (1) {

      if ((stage->interval == 0) && (player_num == 0)) {
	timeout.tv_sec  = 0;
	timeout.tv_usec = 1000; /* ¿Ȥ */
	timeout_p = &timeout;
      } else if (stage->timeout) {
	if ((stage->interval == 0) ||
	    (stage->interval && (stage->timeout_timer < stage->interval))) {
	  timeout.tv_sec  =  stage->timeout_timer / 1000;
	  timeout.tv_usec = (stage->timeout_timer % 1000) * 1000;
	  timeout_p = &timeout;
	}
      }

      ret = 0;
#if 1
      /*
       * Path 饤֥ΥХåե˼ǡĤäƤǽ
       * 뤿ᡤPath 饤֥ФơPathCheckData() ˤ
       * ̤μǡХåե˻ĤäƤʤ䤤碌롥
       *  - 饤 ֤ SEND_KEY ޥɤˤäƱ
       * Τǡ̤μǡХåե˻ĤäƤ뤳ȤϤʤϤ
       * 뤫顤åԤɬפ̵Τå롥
       * (PathCheckData() Ԥ select() ԤƤޤᡤ
       *  Хåե˻Ĥǡ̵ɤȽѤδؿ Path 饤֥
       *  ɲäơäѤ٤ʤΤݤʤΤ PathCheckData()
       *  ѤƤ롥Τᡤ饤ȤΥץ饤ᤤˤϡ
       *  ʲǥեǥץȤ̵̤ select() Ƥޤǽ
       *  )
       */
      FD_ZERO(&fds);
      for (current = ObjList_GetStart(player_list);
	   !ObjList_IsEndEdge(player_list, current);
	   current = ObjListData_GetNext(current)) {
	player = (Player)ObjListData_GetObj(current);
	if (PlayerGetState(player) != PLAYER_ACTIVE) continue;
	if (PathCheckData(PlayerGetPath(player), &command, &length)) {
	  FD_SET(PathGetFileDescriptor(PlayerGetPath(player)), &fds);
	  ret++;
	}
      }
#endif
      if (!ret) {
	memcpy(&fds, &readfds, sizeof(fds));
	ret = select(max_fd + 1, &fds, NULL, NULL, timeout_p);
      }

      if (ret == -1) {
	/*
	 * setitimer() ˤŪ SIGALRM ꡤ
	 * select() ȴ硤⤷ Ctrl-C ʤɤ SIGINT 
	 * ȯ硥
	 */

	/* ʥˤ뽪λ */
	if (IsTerminated()) {
	  break;
	}

	if (player_num == 0) break;

	if (stage->timeout) {
	  if (stage->timeout_timer >= stage->interval)
	    stage->timeout_timer -= stage->interval;
	  else
	    stage->timeout_timer = 0;
	}
      } else if (ret == 0) {
	/* ॢȤξ */
	break;
      } else {
	/* ǡ¸ߤ */

	for (fd = 0; fd <= max_fd; fd++) {
	  if (!FD_ISSET(fd, &fds)) continue;

	  player = fd2player[fd];

	  if (PlayerGetState(player) != PLAYER_ACTIVE) {
	    ErrorExit("Invalid status.");
	  }

	  if (!PathCheckData(PlayerGetPath(player), &command, &length)) {
	    ErrorExit("Invalid path.");
	  }

	  switch (command) {

	  case DELETE_ME :

	    PathRecvNoneData(PlayerGetPath(player), &command);
	    PathSendNoneData(PlayerGetPath(player), DELETE_YOU);
	    PathFlushSendBuffer(PlayerGetPath(player));
	    if (TankGetState(PlayerGetTank(player)) == TANK_ACTIVE)
	      TankSetState(PlayerGetTank(player), TANK_EXPLOSION);
	    PlayerSetState(player, PLAYER_DESTROY);
	    FD_CLR(fd, &readfds);
	    player_num--;
	    break;

	  case NOW_KEY :

	    PathRecvLongData(PlayerGetPath(player), &command, &l1);
	    if (dont_move)
	      l1 = TANK_NO_ACTION;
	    TankSetAction(PlayerGetTank(player), l1);
	    PlayerSetState(player, PLAYER_DONE);
	    FD_CLR(fd, &readfds);
	    player_num--;
	    break;

	  default :
	    ErrorExit("unknown command.");
	  }
	}
      }
    }

    if (fd2player) free(fd2player);
#else
    /* interval 򸫤ơֹ碌 */
    if (stage->interval) {
      while (1) {
	interval_timer = GetTimeForMilliSeconds();
	if (stage->interval_timer > interval_timer)
	  stage->interval_timer -= 256000;
	if (interval_timer - stage->interval_timer >= stage->interval)
	  break;
	usleep((interval_timer - stage->interval_timer) * 10);
      }
      stage->interval_timer = interval_timer;
    } else {
      usleep(10000);
    }

    f1 = 1; /* ץ쥤䡼μ̵ä顤usleep()ԤĤΥե饰 */
    f2 = 1; /* ॢȤνΤΥե饰 */
    do {
      f0 = 0; /* ٤ƤΥץ쥤䡼νѤɤΥե饰 */
      if (!f1) usleep(1000); /* Wait in 1 millisecond. */
      f1 = 0;

      /*
       * νäץ쥤䡼 state  PLAYER_DONE ѹơ
       * ꥹȤƬ˰ư롥
       * Ѥץ쥤䡼ϡꥹȤƬ˻äƤƤΤǡ
       * PLAYER_DONE Ĥä顤ʾ帡ɬפ̵Ȥ
       * Ȥˤʤ롥ΤᡤꥹȤνü鸡롥
       */
      for (current = ObjList_GetEnd(stage->player_list);
	   !ObjList_IsStartEdge(stage->player_list, current);
	   current = ObjListData_GetPrev(current)) {

	player = (Player)ObjListData_GetObj(current);

	/*
	 * Ѥץ쥤䡼ϡꥹȤƬ˻äƤƤΤǡ
	 * PLAYER_DONE Ĥä顤ʾ帡ɬפ̵Ȥ
	 * ȤʤΤǡbreak 롥
	 */
	if (PlayerGetState(player) == PLAYER_DONE) break;

	if (PlayerGetState(player) != PLAYER_ACTIVE) continue;

	if (!PathCheckData(PlayerGetPath(player), &command, &length)) {
	  f0 = 1;
	  continue;
	}
	f1 = 1;

	switch (command) {

	case DELETE_ME :

	  PathRecvNoneData(PlayerGetPath(player), &command);
	  PathSendNoneData(PlayerGetPath(player), DELETE_YOU);
	  PathFlushSendBuffer(PlayerGetPath(player));
	  if (TankGetState(PlayerGetTank(player)) == TANK_ACTIVE)
	    TankSetState(PlayerGetTank(player), TANK_EXPLOSION);
	  PlayerSetState(player, PLAYER_DESTROY);
	  break;

	case NOW_KEY :

	  PathRecvLongData(PlayerGetPath(player), &command, &l1);
	  if (dont_move)
	    l1 = TANK_NO_ACTION;
	  TankSetAction(PlayerGetTank(player), l1);
	  PlayerSetState(player, PLAYER_DONE);

	  /*
	   * Ѥץ쥤䡼ϡꥹȤƬ˻äƤ
	   * ˤꡤʲ롥
	   * PLAYER_ACTIVE ˤʤäƤץ쥤䡼򸡺Τˡ
	   *   ꥹȤ򤹤٤Ƹɬפ̵ʤ롥
	   *   (PLAYER_DONE ˤʤäƤȤޤǸФ褤)
	   * ֿ˻֤ץ쥤䡼ꥹȤƬնˤ뤿ᡤ
	   *   ǡ̤ͥˤʤ롥Ĥޤꡤͥåȥ
	   *   ٤ץ쥤䡼ˤ˥ǡ褦ˤʤ롥
	   *   (餬ŪǤ)
	   */
	  current2 = ObjListData_GetNext(current);
	  ObjList_MoveObjToStart(stage->player_list, current);
	  current = current2;
	  break;

	default :
	  ErrorExit("unknown command.");
	}
      }

      /* ΥॢȽ */
      /* ޤǡƤʤץ쥤䡼Ф */
      if (stage->timeout && f0) {
	if (f2) {
	  /* ॢȤΥޤν */
	  stage->timeout_timer = GetTimeForMilliSeconds();
	  f2 = 0;
	} else {
	  timeout_timer = GetTimeForMilliSeconds();
	  if (stage->timeout_timer > timeout_timer)
	    stage->timeout_timer -= 256000;
	  if (timeout_timer - stage->timeout_timer >= stage->timeout)
	    f0 = 0;
	  /* stage->timeout_timer = timeout_timer; */
	}
      }
    } while (f0);
#endif

    /* Player  State  Done  Active ᤹ */
    for (current = ObjList_GetStart(stage->player_list);
	 !ObjList_IsEndEdge(stage->player_list, current);
	 current = ObjListData_GetNext(current)) {
      player = (Player)ObjListData_GetObj(current);
      switch (PlayerGetState(player)) {
      case PLAYER_WAIT :
	break;
      case PLAYER_ACTIVE :
	/*
	 *  ACTIVE ȤȤϡॢȤȤȤʤΤǡ
	 * ץ쥤䡼롥
	 */
	PathSendNoneData(PlayerGetPath(player), TIMEOUT_YOU);
	PathFlushSendBuffer(PlayerGetPath(player));
	if (TankGetState(PlayerGetTank(player)) == TANK_ACTIVE)
	  TankSetState(PlayerGetTank(player), TANK_EXPLOSION);
	PlayerSetState(player, PLAYER_DESTROY);
	break;
      case PLAYER_DONE :
	PlayerSetState(player, PLAYER_ACTIVE);
	break;
      default :
	break;
      }
    }

    /* ֤ν */
    StageProcTank(stage->map, stage->tank_list, stage->canon_list,
		  stage->missile_list, stage->explosion_list);

    /*
     * ơλȽ
     * (SEND_KEY ׵Ф NOW_KEY ֿƤ饹ơλ
     *  ʤȡNOW_KEY ХåեѤ߽ŤʤäƤơʤ
     *  ȿߤʤȤݤƤޤΤǡNOW_KEY ν
     *  Ȥ˥ơλȽƤ)
     */
    lose = 0;
    for (current = ObjList_GetStart(stage->tank_list);
	 !ObjList_IsEndEdge(stage->tank_list, current);
	 current = ObjListData_GetNext(current)) {
      tank = (Tank)ObjListData_GetObj(current);
      if (TankGetState(tank) == TANK_DESTROY) lose++;
    }
    if (StageIsOver(stage, lose)) break;

    /*
     * ǡ
     */

    MapSendChangesToAllPlayers(stage->map, stage->player_list);

    for (current = ObjList_GetStart(stage->player_list);
	 !ObjList_IsEndEdge(stage->player_list, current);
	 current = ObjListData_GetNext(current)) {

      player = (Player)ObjListData_GetObj(current);
      if (PlayerGetState(player) != PLAYER_ACTIVE) continue;

      PathSendNoneData(PlayerGetPath(player), GAME_CLEAN);

      for (current2 = ObjList_GetStart(stage->tank_list);
	   !ObjList_IsEndEdge(stage->tank_list, current2);
	   current2 = ObjListData_GetNext(current2)) {
	tank = (Tank)ObjListData_GetObj(current2);
	/* if (TankGetState(tank) != TANK_ACTIVE) continue; */ /* ס */
	PathSendCharCharShortShortData(PlayerGetPath(player), TANK_DRAW,
				       (unsigned char)TankGetAngle(tank),
				       (unsigned char)TankGetColorNumber(tank),
				       TankGetX(tank) * 100 + 0.5,
				       TankGetY(tank) * 100 + 0.5);
      }

      for (current2 = ObjList_GetStart(stage->canon_list);
	   !ObjList_IsEndEdge(stage->canon_list, current2);
	   current2 = ObjListData_GetNext(current2)) {
	canon = (Canon)ObjListData_GetObj(current2);
	if (CanonGetState(canon) != CANON_ACTIVE) continue;
	PathSendShortShortData(PlayerGetPath(player), CANON_DRAW,
			       CanonGetX(canon) * 100 + 0.5,
			       CanonGetY(canon) * 100 + 0.5);
      }

      for (current2 = ObjList_GetStart(stage->missile_list);
	   !ObjList_IsEndEdge(stage->missile_list, current2);
	   current2 = ObjListData_GetNext(current2)) {
	missile = (Missile)ObjListData_GetObj(current2);
	if (MissileGetState(missile) != MISSILE_ACTIVE) continue;
	PathSendShortShortData(PlayerGetPath(player), MISSILE_DRAW,
			       MissileGetX(missile) * 100 + 0.5,
			       MissileGetY(missile) * 100 + 0.5);
      }

      for (current2 = ObjList_GetStart(stage->explosion_list);
	   !ObjList_IsEndEdge(stage->explosion_list, current2);
	   current2 = ObjListData_GetNext(current2)) {
	explosion = (Explosion)ObjListData_GetObj(current2);
	if (ExplosionGetState(explosion) != EXPLOSION_ACTIVE) continue;
	PathSendCharShortShortData(PlayerGetPath(player), EXPLOSION_DRAW,
				   ExplosionGetRadius(explosion),
				   ExplosionGetX(explosion) * 100 + 0.5,
				   ExplosionGetY(explosion) * 100 + 0.5);
      }

      PathSendShortShortData(PlayerGetPath(player), DISPLAY_COPY,
			     TankGetX(PlayerGetTank(player)) * 100 + 0.5,
			     TankGetY(PlayerGetTank(player)) * 100 + 0.5);

#if 0 /* ̺̿︺ΤᡤSEND_KEY  XSync() ԤREAD_KEY פˤ */
      PathSendNoneData(PlayerGetPath(player), READ_KEY);
#endif
      PathSendNoneData(PlayerGetPath(player), SEND_KEY);

      PathFlushSendBuffer(PlayerGetPath(player));
    }

    if (dont_move) {
      dont_move_timer = GetTimeForMilliSeconds();
      if (dont_move_timer_start > dont_move_timer)
	dont_move_timer_start -= 256000;
      if (dont_move_timer - dont_move_timer_start >= 3000)
	dont_move = 0;
    }
  }

  if (stage->interval) {
#ifdef USE_ALARM_SIGNAL
    /* 顼ॷʥ̵(setitimer() ̵ɬפꡩ) */
    signal(SIGALRM, SIG_IGN);
#endif
  }

  return (0);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
