/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000-2003 Sakai Hiroaki.                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* Tank 饹(֤Υ饹)Υ饤֥()                       */
/*****************************************************************************/

#ifndef _NADAR_SHARE_TANK_H_INCLUDED_
#define _NADAR_SHARE_TANK_H_INCLUDED_

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

typedef struct _Tank * Tank; /* Tank 饹 */

#include "configure.h"

#include "etc.h"
#include "Obj.h"
#include "Stage.h"
#include "Map.h"
#include "Canon.h"
#include "Missile.h"
#include "Explosion.h"

#define TANK_WAIT      0
#define TANK_ACTIVE    1
#define TANK_EXPLOSION 2
#define TANK_EXPLOSING 3
#define TANK_DESTROY   4

#define TANK_NO_ACTION     0
#define TANK_GO_FORWARD    (1 << 0)
#define TANK_GO_BACKWARD   (1 << 1)
#define TANK_TURN_LEFT     (1 << 2)
#define TANK_TURN_RIGHT    (1 << 3)
#define TANK_FIRE_CANON    (1 << 4)
#define TANK_FIRE_MISSILE  (1 << 5)
#define TANK_DEAD          (1 << 6)

/*===========================================================================*/
/* ѿμ                                                    */
/*===========================================================================*/

double TankGetX(Tank tank);
double TankGetY(Tank tank);
int TankGetAngle(Tank tank);

double TankGetVX(Tank tank);
double TankGetVY(Tank tank);

int TankGetState(Tank tank);
int TankSetState(Tank tank, int state);
unsigned long int TankGetAction(Tank tank);
unsigned long int TankSetAction(Tank tank, unsigned long int action);

int TankGetColorNumber(Tank tank);

Player TankGetPlayer(Tank tank);
Player TankSetPlayer(Tank tank, Player p);

int TankGetCanonNumber(Tank tank);
int TankGetCanonActiveNumber(Tank tank);
int TankGetMissileNumber(Tank tank);
int TankGetMissileActiveNumber(Tank tank);

int TankGetCanonTimer(Tank tank);
int TankGetMissileTimer(Tank tank);

int TankGetDestroyTimer(Tank tank);
int TankSetDestroyTimer(Tank tank, int t);
int TankDecDestroyTimer(Tank tank);

int TankIncCanon(Tank tank);
int TankDecCanon(Tank tank);
int TankIncMissile(Tank tank);
int TankDecMissile(Tank tank);

int TankClearMoved(Tank tank);
int TankGetMoved(Tank tank);

Tank TankGetDestroyed(Tank tank);

/*===========================================================================*/
/* ֥ȤȺ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

Tank TankCreate(double x, double y, int angle, Player player, int state,
		int active, int backward);
Tank TankCreateFromTank(Tank old_tank);

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

Tank TankDestroy(Tank tank);

/*===========================================================================*/
/* ư                                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֤ư                                                                */
/*---------------------------------------------------------------------------*/

int TankAct(Tank tank, Map map, ObjList tank_list,
	    ObjList canon_list, ObjList missile_list, ObjList explosion_list);

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

#endif /* _NADAR_SHARE_TANK_H_INCLUDED_ */

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
